; ACL2 Version 3.1 -- A Computational Logic for Applicative Common
; Lisp Copyright (C) 2006 University of Texas at Austin

; This version of ACL2 is a descendent of ACL2 Version 1.9, Copyright
; (C) 1997 Computational Logic, Inc.  See the documentation topic
; NOTE-2-0.

; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

; Written by:  Matt Kaufmann               and J Strother Moore
; email:       Kaufmann@cs.utexas.edu      and Moore@cs.utexas.edu
; Department of Computer Sciences
; University of Texas at Austin
; Austin, TX 78712-1188 U.S.A.

; The original version of this file was contributed by Bob Boyer and
; Warren A. Hunt, Jr.  The design of this system of Hash CONS,
; function memoization, and fast association lists (applicative hash
; tables) was initially implemented by Boyer and Hunt.

(in-package "ACL2")

;;; Matt, this probably should go into a different file, as Gary
;;; says.

#+openmcl
(let ((*readtable* (copy-readtable nil nil)))
  (in-package "CCL")
  (eval (read-from-string "

   ;;; Work of Gary Byers.

   ;;; The #_ and #$ reader macros in the code below are part of
   ;;; OpenMCL's ffi; you'd basically need to hide this code in
   ;;; a file that's isolated from other implementations.
   (defun acl2::physical-memory ()
      #+darwin-target
      (rlet ((count :mach_msg_type_number_t #$HOST_BASIC_INFO_COUNT)
             (info :host_basic_info))
        (if (eql 0 (#_host_info (#_mach_host_self)
                                #$HOST_BASIC_INFO
                                info
                                count))
          (pref info :host_basic_info.max_mem)))
      #+freebsd-target
       (rlet ((ret :unsigned-doubleword 0)
              (mib (:array :uint 2))
              (oldsize :uint (ccl::record-length
                               :unsigned-doubleword)))
         (setf (paref mib (:* :uint) 0) #$CTL_HW
               (paref mib (:* :uint) 1) #$HW_PHYSMEM)
         (if (eql 0 (#_sysctl mib 2 ret oldsize (%null-ptr) 0))
           (pref ret unsigned-doubleword)))
      #+linux-target
      (rlet ((info :sysinfo))
        (if (eql 0 (#_sysinfo info))
          (pref info :sysinfo.totalram))))"))
  (in-package "ACL2"))

#-openmcl
(defun physical-memory ()
  (cond ((probe-file "/proc/meminfo")
         (let* (n kb
                  (key "MemTotal:")
                  (info (with-open-file (si "/proc/meminfo")
                          (with-output-to-string
                            (so)
                            (let (c)
                              (loop while (setq c (read-char
                                                   si nil nil))
                                    do (write-char c so))))))
                  (loc (search key info)))
           (our-syntax
            (with-input-from-string
             (s info :start (+ loc (length key)))
             (setq n (read s))
             (setq kb (read s))
             (cond ((and (integerp n) (equal kb 'kb))
                    (* n 1024)))))))))

(defmacro free-bytes ()
  #+openmcl
  (ccl::%freebytes)
  #-openmcl
  0)

#+openmcl
(eval-when (load eval compile)
  (if (fboundp 'ccl::rdtsc64) (pushnew :rdtsc64 *features*)))

(defconstant most-positive-fixnum/2 (floor most-positive-fixnum 2))

(defmacro defglobal (&rest r)

; In OpenMCL, DEFSTATIC behaves exactly as DEFPARAMETER; however, its
; use includes three promises: (1) never to locally bind the variable,
; e.g., with LET or LAMBDA, (2) never to reference the variable when
; it is not set, which would be an error anyway, of course, and (3)
; never to test whether the variable is BOUNDP.  OpenMCL uses about
; ten fewer machine instructions to reference such a variable.

  #+openmcl `(ccl::defstatic ,@r)
  #-openmcl `(defparameter ,@r))

(defmacro our-syntax (&rest args)

;  OUR-SYNTAX is like Common Lisp's WITH-STANDARD-IO-SYNTAX.  However,
;  we skip binding *PRINT-MISER-WIDTH*, *PRINT-PPRINT-DISPATCH*, and
;  *READ-DEFAULT-FLOAT-FORMAT* because we do not ever use them.
;  Furthermore, we bind *PACKAGE* and *READTABLE* to their ACL2
;  versions.

;  OUR-SYNTAX is useful when entering into an error break, to be sure
;  things are printed and read in a usual way.  OUR-SYNTAX is also
;  valuable when printing a file one hopes to READ later, e.g., in
;  PRINT-MODULE or COMPACT-PRINT-FILE.

  `(let ((*package*                    *acl2-package*)
         (*print-array*                t)
         (*print-base*                 10)
         (*print-case*                 :upcase)
         (*print-circle*               nil)
         (*print-escape*               t)
         (*print-gensym*               t)
         (*print-length*               nil)
         (*print-level*                nil)
         (*print-lines*                nil)
         (*print-pretty*               nil)
         (*print-radix*                nil)
         (*print-readably*             t)
         (*print-right-margin*         72)
         (*read-base*                  10)
         (*read-eval*                  nil)
         (*read-suppress*              nil)
         (*readtable*                  *acl2-readtable*))
     ,@args))

(defglobal *count-hons-calls*                   t)
(defglobal *count-pons-calls*                   t)
(defglobal *break-honsp*                        nil)
(defglobal *hons-report-discipline-failure*     'break)

(defglobal *report-bytes-used*
  #+openmcl t
  #-openmcl nil)

(defglobal *report-hons-calls* t)

; Gary Byers recalls Lisp folklore that alists are faster than hash
; tables up to length 18.  If one changes these parameters at runtime,
; then it would be best to start afresh with (unmemoize-all) and
; (init-hash-tables).

(defconstant *start-car-ht-size*            18)
(defconstant *hons-acons-ht-threshold*      18)
(defconstant *small-ht-size*                60)
(defconstant *hons-cdr-ht-size*             (expt 2 20))
(defconstant *nil-ht-size*                  (expt 2 17))
(defglobal   *max-memoize-fns*              30)

(declaim (fixnum *start-car-ht-size* *hons-acons-ht-threshold*
                 *small-ht-size* *hons-cdr-ht-size* *nil-ht-size*
                 *max-memoize-fns*))

(defglobal *memoize-debug* nil)

(defglobal *memoize-hack-condition* nil)
(defglobal *memoize-hack-inline* nil)
(defglobal *memoize-hack-trace* nil)

(declaim (fixnum *hons-call-counter* *hons-misses-counter*
                 *pons-call-counter* *pons-hit-counter*))

;  The HONS Invariants

; If A and B are consp+honsp, then (eq A B) iff (equal A B).
; The car of a consp+honsp is an atom or a consp+honsp.
; The cdr of a consp+honsp is an atom or a consp+honsp.
; No consp+honsp is circular.  If a string occurs in any
; consp+honsp, then no other EQUAL string occurs in any
; consp+honsp.

; memoize-flush 'forgets' all that was remembered for certain
; functions that use certain stobjs.  We must keep memoize-flush very
; fast in execution so as not to slow down stojb update or resize
; operations in general.  We 'forget' the pons table later.

(defmacro memoize-flush (st)
  (let ((s (st-lst st)))
    `(loop for sym in ,s do
           (let ((old (symbol-value (the symbol sym))))
             (unless (or (null old) (empty-ht-p old))
               (setf (symbol-value (the symbol sym)) nil))))))

(defmacro mht (&key (test (quote (function eql)))
                    (size *small-ht-size*)
                    (shared nil)
                    (rehash-size 1.5)
                    (rehash-threshold 0.7)
                    (weak nil))
  #-openmcl (declare (ignore shared weak))
  `(make-hash-table :test             ,test
                    :size             (max *small-ht-size* ,size)
                    :rehash-size ,rehash-size
                    :rehash-threshold ,rehash-threshold
                    #+openmcl :weak   #+openmcl ,weak
                    #+openmcl :shared #+openmcl ,shared))

(declaim (hash-table
          *hons-cdr-ht*
          *hons-cdr-ht-eql*
          *nil-ht* *hons-acons-ht*
          *hons-str-ht*
          *memoize-info-ht*
          *memoize-reject-ht*
          *compact-print-file-ht*
          *compact-read-file-ht*))

; Here are some basic data structures for honsing and memoizing.  Some
; of these are significantly expanded in size later by hons-init, but
; there is no reason to clog up saved images with large empty versions
; of them.

; The choice of :weak below deserves careful explanation. !!

(defglobal *hons-cdr-ht*        (mht :test #'eq :weak :key))
(defglobal *hons-cdr-ht-eql*    (mht))
(defglobal *nil-ht*             (mht :weak :value))
(defglobal *hons-acons-ht*      (mht :test #'eq :weak :key))
(defglobal *hons-str-ht*        (mht :test #'equal :weak :value))
(defglobal *hons-copy-aux-ht*   (mht :test #'eq))

(defglobal *memoize-info-ht* (mht))

(defglobal *memoize-array*
  (make-array 1 :element-type 'fixnum :initial-element 0))
(defglobal *memoize-reject-ht*  (mht :test #'eq))

(defmacro str-hash-set (x)
  (check-type x symbol)
  `(setq ,x (maybe-str-hash ,x)))

(defun maybe-str-hash (x)
  (if (typep x '(and array string))
      (or (gethash x *hons-str-ht*)
          (setf (gethash x *hons-str-ht*) x))
    x))

#+cmu
(proclaim `(type (simple-array fixnum (*)) *memoize-array*))
#-cmu
(eval-when
 #-cltl2
 (load eval)
 #+cltl2
 (:load-toplevel :execute)
 (proclaim `(type (simple-array fixnum (*)) *memoize-array*)))

; *float-ticks/second* is set correctly by hons-init.
(defglobal *float-ticks/second* 1.0)
(declaim (float *float-ticks/second*))

(defun total-time (fn)
  (/ (aref *memoize-array*
           (the fixnum (* *max-memoize-fns* (symbol-to-fixnum fn))))
     *float-ticks/second*))

; INTERNAL-REAL-TIME uses the X86 RDTSC instruction when available.
(defmacro internal-real-time ()
  '(the fixnum
     (let ((n #+RDTSC64 (ccl::rdtsc64)
              #-RDTSC64 (get-internal-real-time)))
       (if (typep n 'fixnum)
           n
         #+RDTSC64
         (error "~&; Error: ** rdtsc64 returned a nonfixnum.")
         #-RDTSC64
         (error "~&; Error: ** get-internal-real-time returned a ~
                 nonfixnum.")))))

(defmacro maybe-break-honsp ()
  (cond (*break-honsp* `(progn (break "~&; HONSP returned nil.")))))

(defmacro honsp (x)

; HONSP checks a cons see if it's in our hons tables.
; HONSP assumes x is a CONSP.

  `(cond ((let* ((x ,x)
                 (ax (car x))
                 (dx (cdr x))
                 (v (cond ((null dx) *nil-ht*)
                          ((consp dx) (gethash dx *hons-cdr-ht*))
                          (t (gethash dx *hons-cdr-ht-eql*)))))
            (if (listp v)
                (let ((av (car v)))
                  (if (typep ax '(or cons symbol (and array string)))
                      (loop (if (eq ax (car av)) (return (eq x av)))
                            (setq v (cdr v))
                            (if (null v) (return nil))
                            (setq av (car v)))
                    (loop (if (eql ax (car av)) (return (eq x av)))
                          (setq v (cdr v))
                          (if (null v) (return nil))
                          (setq av (car v)))))
              (eq x (gethash (car x) v)))))
         (t (maybe-break-honsp) nil)))

(defglobal *hons-call-counter* 0)
(defglobal *hons-misses-counter* 0)
(defglobal *pons-call-counter* 0)
(defglobal *pons-hit-counter* 0)

(defun safe-incf-error ()
  (ofe "~&; safe-incf:  ** Error:  overflow."))

(defmacro safe-incf (x &optional (inc 1))
  ; both x and inc should be nonnegative fixnums
  ; <= than most-positive-fixnum/2 so that their sum
  ; is a fixnum, too.
  (let ((v (gensym "V")))
    (cond ((eql inc 1)
           `(let ((,v ,x))
              (declare (fixnum ,v))
              (if (eql ,v most-positive-fixnum)
                  (safe-incf-error)
                (the fixnum (setf ,x (the fixnum (+ 1 ,v)))))))
          (t (let ((iv (gensym "IV")))
               `(let ((,v ,x)
                      (,iv ,inc))
                  (declare (fixnum ,v ,iv))
                  (cond ((and (<= ,v most-positive-fixnum/2)
                              (<= ,iv most-positive-fixnum/2))
                         (the fixnum
                           (setf ,x (the fixnum (+ ,v ,iv)))))
                        (t (safe-incf-error)))))))))

(defmacro maybe-count-hons-calls ()
  (and *count-hons-calls* '(safe-incf *hons-call-counter*)))
(defmacro maybe-count-hons-misses ()
  (and *count-hons-calls* '(safe-incf *hons-misses-counter*)))
(defmacro maybe-count-pons-calls ()
  (and *count-pons-calls* '(safe-incf *pons-call-counter*)))
(defmacro maybe-count-pons-hits ()
  (and *count-pons-calls* '(safe-incf *pons-hit-counter*)))

(defun assoc-no-error-at-end (x l)

; We assume that every element of L is CONSP.

  (if (typep x '(or cons symbol (and array string)))
      (loop (if (consp l)
                (let ((al (car l)))
                  (if (eq x (car al))
                      (return al)
                    (setq l (cdr l))))
              (return nil)))
    (loop (if (consp l)
              (let ((al (car l)))
                (if (eql x (car al))
                    (return al)
                  (setq l (cdr l))))
            (return nil)))))

(defun too-long (x n)
  (declare (fixnum n))
  (loop (cond ((atom x) (return nil))
              ((eql n 0) (return t))
              (t (setq x (cdr x))
                 (setq n (the fixnum (1- n)))))))

; HONS

; Definition of NORMED.  An ACL2 Common Lisp object x is normed iff
; both (a) if x is a string, then it is in *HONS-STR-HT* and (b) if x
; is a CONSP, then it is also a HONSP.  HONS-COPY and HONS produce
; normed objects.  A non-HONSP cons whose car and cdr are normed is
; honsed up by putting it in the right place in the hash tables, after
; first checking that it can be legitimately placed there.

(defun hons-normed-with-suggestion (x y nhons)
  (let* ((yt (if (consp y) *hons-cdr-ht* *hons-cdr-ht-eql*))
         (yval (if y (gethash y yt) *nil-ht*))
         (yp (listp yval)))
    (maybe-count-hons-calls)
    (cond ((if yp (assoc-no-error-at-end x yval) (gethash x yval)))
          (t (maybe-count-hons-misses)
             (cond (yp
                    (cond ((too-long yval *start-car-ht-size*)
                           (let ((tab (mht :weak :value)))
                             (loop for pair in yval do
                                   (setf (gethash (car pair) tab)
                                         pair))
                             (setf (gethash (car nhons) tab) nhons)
                             (setf (gethash y yt) tab)
                             nhons))
                          (t (setf (gethash y yt) (cons nhons yval))
                             nhons)))
                   (t (setf (gethash x yval) nhons)))))))

(defun hons-normed (x y)
  (let* ((yt (if (consp y) *hons-cdr-ht* *hons-cdr-ht-eql*))
         (yval (if y (gethash y yt) *nil-ht*))
         (yp (listp yval)))
    (maybe-count-hons-calls)
    (cond ((if yp (assoc-no-error-at-end x yval) (gethash x yval)))
          (t (maybe-count-hons-misses)
             (let ((nhons (cons x y)))
               (cond
                (yp
                 (cond ((too-long yval *start-car-ht-size*)
                        (let ((tab (mht :weak :value)))
                          (loop for pair in yval do
                                (setf (gethash (car pair) tab)
                                      pair))
                          (setf (gethash (car nhons) tab) nhons)
                          (setf (gethash y yt) tab)
                          nhons))
                       (t (setf (gethash y yt) (cons nhons yval))
                          nhons)))
                (t (setf (gethash x yval) nhons))))))))

; HONS-COPY

; In general, HONS-COPY has no justification for reusing, much less
; smashing, the conses it is passed.  However, during
; CLEAR-HASH-TABLES, in the rehons phase, such reuse is precisely what
; is needed and permitted.  If some day we learn that no cons in, say,
; (w state) will ever be RPLACAed or RPLACDed and that REPLACAing any
; cons in it with an EQUAL CAR value is ok, and same for RPLACD/CDR,
; then we could probably legitimately absorb the conses in (w state)
; as honses via HONS-COPY1-CONSUME.  HONS-COPY is partially and
; temporarily self-memoizing.

(defun uncopy (x)
  (unless (atom x)
    (when (remhash x *hons-copy-aux-ht*)
      (uncopy (car x))
      (uncopy (cdr x)))))

(defabbrev hons-copy2-consume (x)
  (let ((a (hons-copy3-consume (car x)))
        (d (hons-copy3-consume (cdr x))))
    (or (eql a (car x)) (rplaca x a))
    (or (eql d (cdr x)) (rplacd x d))
    (hons-normed-with-suggestion a d x)))

(defun hons-copy1-consume (x)
  (cond ((atom x) (maybe-str-hash x))
        ((honsp x) x)
        (t (unless (eql 0 (hash-table-count *hons-copy-aux-ht*))
             ; Only get here because of an error during a HONS-COPY.
             (setq *hons-copy-aux-ht* (mht :test #'eq)))
           (let ((ans (hons-copy2-consume x)))
             (unless (eql 0 (hash-table-count *hons-copy-aux-ht*))
               (uncopy (car ans))
               (uncopy (cdr ans)))
             ans))))

(defun hons-copy3-consume (x)
  (cond ((atom x) (maybe-str-hash x))
        ((honsp x) x)
        ((gethash x *hons-copy-aux-ht*))
        (t (setf (gethash x *hons-copy-aux-ht*)
                 (hons-copy2-consume x)))))

(defabbrev hons-copy2 (x)
  (let ((a (hons-copy3 (car x)))
        (d (hons-copy3 (cdr x))))
    (hons-normed a d)))

(defun hons-copy3 (x)
  (cond ((atom x) (maybe-str-hash x))
        ((honsp x) x)
        ((gethash x *hons-copy-aux-ht*))
        (t (setf (gethash x *hons-copy-aux-ht*)
                 (hons-copy2 x)))))

(defun hons-copy1 (x)
  (cond ((atom x) (maybe-str-hash x))
        ((honsp x) x)
        (t (unless (eql 0 (hash-table-count *hons-copy-aux-ht*))
             ; Only get here because of an error during a hons-copy.
             (setq *hons-copy-aux-ht* (mht :test #'eq)))
           (let ((ans (hons-copy2 x)))
             (unless (eql 0 (hash-table-count *hons-copy-aux-ht*))
               (uncopy (car ans))
               (uncopy (cdr ans)))
             ans))))

(defun hons-copy (x) ; ACL2 user-visible.
  (hons-copy1 x))

(defun hons-when-x-is-honsp (x y)
  (hons-normed x (hons-copy1 y)))

(defun hons-when-y-is-honsp (x y)
  (hons-normed (hons-copy1 x) y))

; HONS-EQUAL is ACL2 user-visible, identical in ACL2 logical meaning
; to EQUAL.
(defun hons-equal (x y)
  (cond ((eq x y))
        ((atom x)
         (cond ((consp y) nil)
               ((symbolp y) nil)
               ((typep y 'fixnum) (eql x y))
               (t (equal x y))))
        ((atom y) nil)
        ((honsp x)
         (cond ((honsp y) nil)
               (t (and (hons-equal-h1 (car x) (car y))
                       (hons-equal-h1 (cdr x) (cdr y))))))
        ((honsp y)
         (and (hons-equal-h1 (car y) (car x))
              (hons-equal-h1 (cdr y) (cdr x))))
        (t (and (hons-equal (car y) (car x))
                (hons-equal (cdr y) (cdr x))))))

; HONS-EQUAL-H1 is like HONS-EQUAL, but with the assumption that X has
; been normed.
(defun hons-equal-h1 (x y)
  (cond ((eq x y))
        ((atom x)
         (cond ((consp y) nil)
               ((symbolp y) nil)
               ((typep y 'fixnum) (eql x y))
               (t (equal x y))))
        ((atom y) nil)
        ((honsp y) nil)
        (t (and (hons-equal-h1 (car x) (car y))
                (hons-equal-h1 (cdr x) (cdr y))))))

; HONS is ACL2 user-visibl and is identical in ACL2 logical meaning to
; CONS.
(defun hons (x y)
  (hons-normed (hons-copy1 x) (hons-copy1 y)))

; HONS-GET, HONS-ACONS and HONS-ACONS!

; HONS-ACONS and HONS-GET provide fast lookup in alists, with
; ASSOC-EQUAL semantics but with the speed of hash tables in some
; cases.  These operations permit one to reasonably efficiently work
; with extremely long alists in some cases.  Informally speaking, each
; HONS-ACONS operation steals the hash table associated with the alist
; that is being extended.  The key is always honsed before the hashing
; operation in HONS-ACONS, and HONS-ACONS!.  In order to take
; advantage of the hash table lookup speed, when coding one must be
; very conscious of which object is the most recent extension of an
; alist and use that extension exclusively.  This may require careful
; passing of the alist up and down through function calls, as with any
; single threaded object in an applicative setting.  There is no
; syntactic enforcement to force one to only use the most recent
; extension of an alist, as there is for single threaded objects.  The
; only penalty for a failure to keep track of the most recent
; extension is a loss of execution speed, not of correctness.  And
; perhaps the annoyance of some warning messages about 'discipline'.

; If we limit ourselves to alists that are recognized by ALISTP, a
; possible gotcha when using HONS-ACONS! is that it might "steal" a
; hash table without one's expecting it.  For example, if you start
; two alists with (HONS-ACONS! '1 '2 nil) and (HONS-ACONS! '1 '2 nil),
; then adding something to the first with HONS-ACONS! will "steal" the
; table associated with what you thought was the second, with the
; result that adding things to the second will result in slow access.
; One can get around this annoyance to some extent by putting
; something 'unique' at the end of the alist, e.g., the final cdr.
; Our (partial) fix for this is to permit the final NIL of an
; HONS-GET/HONS-ACONS! association list to be any symbol, effectively
; naming each association list and maybe preventing the "gotcha" just
; mentioned.

(defmacro maybe-report-discipline-failure (fn)
  (cond (*hons-report-discipline-failure*
         `(cond ((eq *hons-report-discipline-failure* 't)
                 (format *debug-io*
                         "~&; Warning: ~a discipline failure.~%" ',fn)
                 nil)
                ((eq *hons-report-discipline-failure* 'break)
                 (break "~&; Break: ~a discipline failure.~%" ',fn)
                 nil)))))

(defun-one-output hons-get-fn-do-not-hopy (key l)
  (when (atom l) (return-from hons-get-fn-do-not-hopy nil))
  (let (h)
    (when (or (and (consp key) (not (honsp key)))
              (null (setq h (gethash l *hons-acons-ht*))))
      (return-from hons-get-fn-do-not-hopy (hons-assoc-equal key l)))
    (let ((key (hons-copy1 key)))
      (loop
       (cond ((typep h 'fixnum)
              (return (assoc-no-error-at-end key l)))
             (h (return (values (gethash key h))))
             (t (cond ((and (consp (car l))
                            (hons-equal-h1 key (caar l)))
                       (return (car l)))
                      (t (setq l (cdr l))
                         (when (atom l) (return nil))
                         (setq h (gethash (cdr l)
                                          *hons-acons-ht*))))))))))

(defun-one-output hons-get-fn-do-hopy (key l)
  (when (atom l) (return-from hons-get-fn-do-hopy nil))
  (let ((h (gethash l *hons-acons-ht*)))
    (when (null h)
      (maybe-report-discipline-failure 'hons-get-fn-do-hopy))
    (let ((key (hons-copy1 key)))
      (loop
       (cond ((typep h 'fixnum)
              (return (assoc-no-error-at-end key l)))
             (h (return (values (gethash key h))))
             (t (cond ((and (consp (car l))
                            (hons-equal-h1 key (caar l)))
                       (return (car l)))
                      (t (setq l (cdr l))
                         (when (atom l) (return nil))
                         (setq h (gethash (cdr l)
                                          *hons-acons-ht*))))))))))

; Why do we want both HONS-ACONS and HONS-ACONS!, which is the HONSing
; version of HONS-ACONS?  On the one hand, since it is quite possible
; that one will not want to look further into the alist that is
; formed, given fast hashing lookup, one may not wish the overhead of
; HONSing it.  On the other hand, if the alist is going to be an
; argument to a function that is to be memoized, then the memoization
; process may hons it -- possibly over and over and over, which can be
; very time consuming if it is very long.

(defun hons-acons (key value l)
  (setq key (hons-copy1 key))
  (let ((ans (cons (cons key value) l)))
    (cond ((atom l)
           (setf (gethash ans *hons-acons-ht*) 0))
          (t (let ((tab (gethash l *hons-acons-ht*)))
               (remhash l *hons-acons-ht*)
               (cond
                ((typep tab 'fixnum)
                 (cond ((< (the fixnum tab) *hons-acons-ht-threshold*)
                        (setf (gethash ans *hons-acons-ht*)
                              (the fixnum (+ 1 tab))))
                       (t (let ((tab (mht :test #'eq :weak :key)))
                            (loop for tail on ans
                                  unless (gethash (caar tail) tab)
                                  do (setf (gethash (caar tail) tab)
                                           (car tail)))
                            (setf (gethash ans *hons-acons-ht*)
                                  tab)))))
                (tab
                 (setf (gethash key tab) (car ans))
                 (setf (gethash ans *hons-acons-ht*) tab))
                (t (maybe-report-discipline-failure 'hons-acons))))))
    ans))

(defun hons-acons! (key value l)
  (setq key (hons-copy1 key))
  (let ((ans (hons-when-x-is-honsp
              (hons-when-x-is-honsp key value)
              l)))
    (cond ((atom l)
           (setf (gethash ans *hons-acons-ht*) 0))
          (t (let ((tab (gethash l *hons-acons-ht*)))
               (remhash l *hons-acons-ht*)
               (cond
                ((typep tab 'fixnum)
                 (cond ((< (the fixnum tab) *hons-acons-ht-threshold*)
                        (setf (gethash ans *hons-acons-ht*)
                              (the fixnum (+ 1 tab))))
                       (t (let ((tab (mht :test #'eq :weak :key)))
                            (loop for tail on ans
                                  unless (gethash (caar tail) tab)
                                  do (setf (gethash (caar tail) tab)
                                           (car tail)))
                            (setf (gethash ans *hons-acons-ht*)
                                  tab)))))
                (tab
                 (setf (gethash key tab) (car ans))
                 (setf (gethash ans *hons-acons-ht*) tab))
                (t (maybe-report-discipline-failure 'hons-acons!))))))
    ans))

(defun hons-shrink-alist-orig (alcdr ans)
  (cond ((atom alcdr) ans)
        (t (let ((p (hons-get (car (car alcdr)) ans)))
             (cond (p (hons-shrink-alist-orig (cdr alcdr) ans))
                   (t (hons-shrink-alist-orig
                       (cdr alcdr)
                       (hons-acons (car (car alcdr))
                                   (cdr (car alcdr))
                                   ans))))))))

(defun hons-shrink-alist-help (alcdr ans tab)
  (if (atom alcdr)
      (progn (setf (gethash ans *hons-acons-ht*) tab)
             ans)
    (let* ((pair (car alcdr))
           ;; (key (hons-copy (car pair)))
           (key (car pair)) ; We know (car pair) is HONS-NORMEDP
           (val (gethash key tab))
           (ans (if val ans (cons (cons key (cdr pair)) ans)))
           (tab (if val tab
                  (progn
                    (setf (gethash key tab) (car ans))
                    tab))))
      (hons-shrink-alist-help (cdr alcdr) ans tab))))

(defun hons-shrink-alist (alcdr ans)
  (if (atom alcdr)
      ans
    (let* ((tab (gethash alcdr *hons-acons-ht*))
           (ans-size
            (if (hash-table-p tab)
                #-openmcl (1+ (ceiling
                               (hash-table-count tab)
                               .7))
                #+openmcl (1+ (hash-table-count tab))
                nil)))
      (if (or (not ans-size) (consp ans))
          (hons-shrink-alist-orig alcdr ans)
        (let ((ans-tab (mht :test #'eq :size ans-size :weak :key)))
          (hons-shrink-alist-help alcdr ans ans-tab))))))

(defun hons-shrink-alist!-orig (alcdr ans)
  (cond ((atom alcdr) ans)
        (t (let ((p (hons-get (car (car alcdr)) ans)))
             (cond (p (hons-shrink-alist!-orig (cdr alcdr) ans))
                   (t (hons-shrink-alist!-orig
                       (cdr alcdr)
                       (hons-acons! (car (car alcdr))
                                    (cdr (car alcdr))
                                    ans))))))))

(defun hons-shrink-alist!-help (alcdr ans tab)
  (if (atom alcdr)
      (progn (setf (gethash ans *hons-acons-ht*) tab)
             ans)
    (let* ((pair (car alcdr))
           ;; (key (hons-copy (car pair)))
           (key (car pair))
           (val (gethash key tab))
           (ans (if val ans (hons-normed
                             (hons-when-x-is-honsp key (cdr pair))
                             ans)))
           (tab (if val tab
                  (progn
                    (setf (gethash key tab) (car ans))
                    tab))))
      (hons-shrink-alist!-help (cdr alcdr) ans tab))))

(defun hons-shrink-alist!-help-honsp-alcdr (alcdr ans tab)
  (if (atom alcdr)
      (progn (setf (gethash ans *hons-acons-ht*) tab)
             ans)
    (let* ((pair (car alcdr))
           ;; (key (hons-copy (car pair)))
           (key (car pair))
           (val (gethash key tab))
           (ans (if val ans (hons-normed pair ans)))
           (tab (if val tab
                  (progn
                    (setf (gethash key tab) (car ans))
                    tab))))
      (hons-shrink-alist!-help-honsp-alcdr (cdr alcdr) ans tab))))

(defun hons-shrink-alist! (alcdr ans)
  (if (atom alcdr)
      ans
    (let* ((ans (maybe-str-hash ans))
           (tab (gethash alcdr *hons-acons-ht*))
           (ans-size
            (if (hash-table-p tab)
                #-openmcl (1+ (ceiling
                               (hash-table-count tab)
                               .7))
                #+openmcl (1+ (hash-table-count tab))
                nil)))
      (if (or (not ans-size) (consp ans))
          (hons-shrink-alist!-orig alcdr ans)
        (let ((ans-tab (mht :test #'eq :size ans-size :weak :key)))
          (if (honsp alcdr)
              (hons-shrink-alist!-help-honsp-alcdr alcdr ans ans-tab)
            (hons-shrink-alist!-help alcdr ans ans-tab)))))))

; MEMOIZE

; *memoize-info-ht* maps each currently memoized function symbol to a
; list:

; tablename      a symbol whose value is the memoize table or is nil
; ponstablename  a symbol whose value is the the pons table or is nil
; old-fn         a symbol whose value is the original definition
; memoized-fn    a symbol whose value is the new (memoized) function
; inline         T or NIL
; condition      the condition function (which can be T or NIL)
; num            a nonnegative integer, unique to the function
; sts            the stobj memotable lists for this function
; trace          T or NIL
; start-time     a symbol with val time of the current call or -1

; *memoize-info-ht* also maps num back to the corresponding symbol.

(defrec memoize-info-ht-entry
  (tablename ponstablename old-fn memoized-fn
             condition inline num sts trace start-time)
  nil)

(defun memoize-array-grow ()
  (let* ((nmax (floor (* 1.5 *max-memoize-fns*))))
    (unless (< (* nmax nmax) most-positive-fixnum)
      (ofe "~&; memoize-array-grow: ** Error: most-positive-fixnum ~
            exceeded."))
    (when (> (* nmax nmax) array-total-size-limit)
      (ofe "~&; memoize-array-grow: ** Error: ARRAY-TOTAL-SIZE-LIMIT ~
            exceeded.  Too many memoized functions."))
    (let ((nmax nmax)
          (narray
           (make-array (* nmax nmax)
                       :element-type 'fixnum
                       :initial-element 0)))
      (declare (fixnum nmax)
               (type (simple-array fixnum (*)) narray))
      (loop for i fixnum below *max-memoize-fns* do
            (let ((i1 (* i *max-memoize-fns*))
                  (i2 (* i nmax)))
              (declare (fixnum i1 i2))
              (loop for j fixnum below *max-memoize-fns* do
                    (let ((x (aref *memoize-array*
                                   (the fixnum (+ i1 j)))))
                      (unless (eql x 0)
                        (setf (aref narray
                                    (the fixnum (+ i2 j)))
                              x))))))
      (setq *max-memoize-fns* nmax *memoize-array* narray))))

; SYMBOL-TO-FIXNUM-CREATE

(defglobal *max-symbol-to-fixnum* 4)
(declaim (fixnum *max-symbol-to-fixnum*))

; In SYMBOL-TO-FIXNUM-CREATE, symbol numbers will start at 5 because
; we use earlier slots as follows:

(defconstant *ma-runtime-index*     0)
(defconstant *ma-hits-index*        1)
(defconstant *ma-mht-index*         2)
(defconstant *ma-hons-calls-index*  3)
(defconstant *ma-bytes-used-index*  4)

(defun symbol-to-fixnum-create (s)
  (check-type s symbol)
  (let ((g (gethash s *memoize-info-ht*)))
    (if g (access memoize-info-ht-entry g :num)
      (let (new)
        (loop
         (loop for i from 5 below *max-memoize-fns* do
               (cond ((not (gethash i *memoize-info-ht*))
                      (setq new i)
                      (return nil))))
         (cond (new
                (setq *max-symbol-to-fixnum*
                      (max *max-symbol-to-fixnum* new))
                (return new))
               (t (let ((i *max-memoize-fns*))
                    (memoize-array-grow)
                    (setq *max-symbol-to-fixnum* i)
                    (return-from symbol-to-fixnum-create i)))))))))

(defun symbol-to-fixnum (s)
  (check-type s symbol)
  (let ((g (gethash s *memoize-info-ht*)))
    (if g (access memoize-info-ht-entry g :num)
      (ofe "~&; symbol-to-fixnum:  ** Error:  illegal symbol:  ~s."
           s))))

(defun fixnum-to-symbol (n)
  (check-type n fixnum)
  (or (gethash n *memoize-info-ht*)
      (ofe "~&; fixnum-to-symbol:  ** Error:  illegal number:  ~d."
           n)))

; This code has the 'feature' that if the condition causes an error,
; so will the memoized function.

(defmacro pist (table &rest x)
  (cond ((atom x) nil)
        (t (list 'pons (car x)
                 (cons 'pist (cdr x)) table))))

(defmacro pist* (table &rest x)
  (cond ((atom x) x)
        ((atom (cdr x)) (car x))
        (t (list 'pons (car x)
                 (cons 'pist* (cons table (cdr x))) table))))

; PONS differs from HONS in that it does not honsify its arguments and
; in that it takes a hash table as a third argument.  We use PONS in
; memoization.

; We use PONS instead of HONS in memoization because we could not
; afford to honsify (using hons-shrink-alist!) certain alists in
; certain biology tests.  About the same time, we (gratuitously)
; decided to stop hons'ifying the output of memoized functions.

(defun pons (x y ht)
  (let ((xval nil)
        (yval nil)
        (ans nil))
    (maybe-count-pons-calls)

; We have taken string normalization out of pons because there might
; be a chance of confusing a 'normal' string with a stobj.

; If x1, ..., xn is pointwise EQL to y1, ..., yn, then are we sure
; that (pist* x1 ... xn) is EQ to (pist* y1 ... yn)?

; If CONS exists, then return it.  Does CDR exist in hash table?

    (setq yval (gethash y ht))

; Does CAR exist in hash table?
    (cond (yval
           (cond ((not (consp yval))
                  (setq xval (gethash x yval))
                  (cond (xval
                         (maybe-count-pons-hits)
                         (setq ans xval))))
                 ((setq ans (assoc-no-error-at-end x yval))
                  (maybe-count-pons-hits)))))
    (cond
; If PONS found, then return previous CONS from hash table.
     (ans)
; Otherwise, maybe create new CONS and install in hash table.
     (t
; Make sure that PONS arguments X and Y are both in the hash
; table.  If not, recursively copy X and Y into hash table.
; Find place for CDR.
      (setq yval (gethash y ht))
; Find place for CAR, and reuse CONS if possible.
      (cond
       ((hash-table-p yval)
        (setq xval (gethash x yval))
        (cond ((not xval)
               (setf (gethash x yval)
                     (setq ans (cons x y))))
              (t (maybe-count-pons-hits)
                 (setq ans xval)))
        ans)
       ((not yval)
        (setq ans (cons x y))
        (setf (gethash y ht)
              (list ans))
        ans)
; Create new CONS and put in hash table.  Remember ASSOC
; used for CDR table entries with less than 10 entries.
       (t (let ((ans (assoc-no-error-at-end x yval)))
            (cond
             (ans
              (maybe-count-pons-hits)
              ans)
             (t (let ((ans (cons (cons x y) yval)))
                  (cond
                   ((too-long ans *start-car-ht-size*)
                    (let ((tab (mht)))
                      (loop for pair in ans do
                            (setf (gethash (car pair) tab) pair))
                      (setf (gethash y ht) tab)
                      (car ans)))
                   (t (setf (gethash y ht) ans)
                      (car ans)))))))))))))

(defun memoize-eval-compile (defun)
  (let* (#+gcl
         (system:*notify-gbc* nil)
         #+gcl
         (*load-verbose* nil)
         #+openmcl
         (ccl:*warn-if-redefine-kernel* nil))
    (multiple-value-bind (ans warn-p error-p)
        (compile (eval defun))
      (cond ((or warn-p error-p)
             (ofe "~&; memoize-eval-compile: **  Problem in ~
                    compilation of~%~s,"
                  (list :defun defun
                        :values-returned-by-compile
                        (list ans warn-p error-p)))))
      ans)))

(defun memoizedp-raw (fn)
  (and (symbolp fn)
       (values (gethash fn *memoize-info-ht*))))

(defmacro ofd (&rest r) ; For warnings.
  `(format *debug-io* ,@r))

(defmacro ofn (&rest r) ; For forming strings.
  `(format nil ,@r))

(defun numstr (n)
  (if (typep n '(integer 0 999))
      (format nil "~d" n)
    (format nil "~,1e" n)))

(defmacro ofni (&rest r)
  `(intern (format nil ,@r) "ACL2"))

(defmacro oft (&rest r) ; For *standard-output*.
  `(format t ,@r))

(defmacro oftr (&rest r) ; Fyi.
  `(format *trace-output* ,@r))

(defun ofe (&rest r)  ; Causes an error.
  (our-syntax
   (apply #'format *error-output* r)
   (error "")))

(defun suffix (str sym)
  (check-type str string)
  (check-type sym symbol)
  (let ((spkn (package-name (symbol-package sym)))
        (sn (symbol-name sym)))
    (ofn "~s,~s,~s" str spkn sn)))

; HONS-GENTEMP
(defglobal *hons-gentemp-counter* 0)
(declaim (integer *hons-gentemp-counter*))
(defun-one-output hons-gentemp (root)
  (check-type root string)
  (loop
   (safe-incf *hons-gentemp-counter*)
   (let ((name (ofn "HONS-G-~s,~s" root *hons-gentemp-counter*)))
     (multiple-value-bind (sym status)
         (intern name (find-package "ACL2_INVISIBLE"))
       (if (null status) (return sym))))))

; ST-LST returns a symbol whose value is a list in which are saved the
; names of the memoize tables that will be set to nil whenever the
; stobj st is changed.
(defun st-lst (st)
  (check-type st symbol)
  (multiple-value-bind (symbol status)
      (intern (ofn "HONS-S-~s,~s"
                   (package-name (symbol-package st))
                   (symbol-name st))
              (find-package "ACL2_INVISIBLE"))
    (or status (eval `(defglobal ,symbol nil)))
    symbol))

(defun dcls (l)
     (loop for dec in l nconc
           (let ((temp
                  (if (consp dec)
                      (loop for d in (cdr dec) nconc
                            (if (and (consp d) (eq (car d) 'ignore))
                                nil
                              (cons d nil))))))
             (if temp (cons (cons 'declare temp) nil)))))

(defun timer-error ()
  (ofe "~&; timer-error:  ** Error."))

; MEMOIZE-FN

(defglobal *memoize-fn-cl-error-msg*
 "
 ; Memoizing a function in the COMMON-LISP package is a violation of
 ; the rules of Common Lisp, and consequently this ACL2 session is
 ; unsound.  It is ok to do such memoizing for experimental reasons,
 ; just as one might trace such a function, which, strictly speaking,
 ; is also forbidden.")

; *CALLER* is bound by memoized functions, so we use DEFPARAMETER
; rather than DEFGLOBAL.
(defparameter *caller* 0)
(declaim (fixnum *caller*))

; The :CONDITION parameter of MEMOIZE-FN can either be T, or a
; function symbol defined by the user within the ACL2 loop, or a LISTP
; (CONSP or NIL).  In the last case we think of the condition as an
; expression in the formals of FN.  If the :INLINE parameter T, then
; the body of FN is placed inline in the memoized definition;
; otherwise, a funcall of the original function is placed there.

(defglobal *memoize-init-done* nil)
(defparameter *in-memoize* nil)

(defglobal *last-gc-exit-time* -1)
(declaim (fixnum *last-gc-exit-time*))

(defun memoize-fn
  (fn &key
   (condition  t)
   (inline     t)
   (trace     nil)
   (cl-defun   nil cl-defun-supplied-p)
   (formals    t   formals-supplied-p)
   (stobjs-in  t   stobjs-in-supplied-p)
   (stobjs-out t   stobjs-out-supplied-p))
 (with-warnings-suppressed
  (unless *memoize-init-done*
    (ofe "~%; Memoize-fn:  ** Error: *memoize-init-done* is nil."))
  (unless (and (symbolp fn) (fboundp fn))
    (ofe "~%; Memoize-fn:  ** Error: ~a is not an fboundp symbol."
         fn))
  (when (memoizedp-raw fn)
    (ofe "~%; Memoize-fn:  ** Error:  ~a is already memoized." fn))
  (when (macro-function fn)
    (ofe "~%; Memoize-fn:  ** Error:  ~a is a macro." fn))
  (when (special-form-or-op-p fn)
    (ofe "~%; Memoize-fn:  ** Error:  ~a is special." fn))
  (when (gethash fn *memoize-reject-ht*)
    (ofe "~%; Memoize-fn:  ** Error:  ~a is used in memoization."
         fn))
  (let*
   ((*in-memoize* t)
    (state *the-live-state*)
    (w (w state))
    (cl-defun (cond (cl-defun-supplied-p cl-defun)
                    ((null inline) nil)
                    (t (cltl-def-from-name fn nil w))))
    (formals (if formals-supplied-p formals
               (getprop fn 'formals t 'current-acl2-world w)))
    (stobjs-in (if stobjs-in-supplied-p stobjs-in
                 (getprop fn 'stobjs-in t 'current-acl2-world w)))
    (stobjs-out (if stobjs-out-supplied-p stobjs-out
                  (getprop fn 'stobjs-out t 'current-acl2-world w))))
   (when (and (null cl-defun) inline)
     (ofe "~%; Memoize-fn:  ** Error:  ~a lacks a cltl-def." fn))
   (when (eq t formals)
     (ofe "~%; Memoize-fn:  ** Error: ~a lacks a formals property."
          fn))
   (when (eq t stobjs-in)
     (ofe "~%; Memoize-fn:  ** Error: ~a lacks a stobjs-in property."
          fn))
   (when (eq t stobjs-out)
     (ofe "~%; Memoize-fn: ** Error: ~a lacks a stobjs-out property."
          fn))
   (when (or (member 'state stobjs-in) (member 'state stobjs-out))
     (ofe "~%; Memoize-fn: ** Error:  ~a uses STATE." fn))
   (when (eq (symbol-package fn) *main-lisp-package*)
     (ofd "~&; Memoize-fn:  A warning about the memoization of ~a."
          fn)
     (ofd *memoize-fn-cl-error-msg*)
     (f-put-global 'certify-book-disabledp t state))
   (when (member fn (trace))
     (ofd "~%; Memoize-fn:  Untracing ~a before memoizing it." fn)
     (eval `(untrace ,fn)))
  (let*
   ((*acl2-unwind-protect-stack*
     (cons nil *acl2-unwind-protect-stack*))
    (old-fn (eval `(defglobal ,(hons-gentemp (suffix "ORIG-DEF-" fn))
                     ,(symbol-function fn))))
    (body
     (if inline (car (last cl-defun)) `(funcall ,old-fn ,@formals)))
    (condition-body
     (cond ((or (eq condition t) (eq condition nil)) condition)
           ((symbolp condition)
            (car (last (cltl-def-from-name condition nil w))))
           (t condition)))
    (dcls (dcls (cdddr (butlast cl-defun))))
    (fnn (symbol-to-fixnum-create fn))
    (start-time
     (let ((v (hons-gentemp (suffix "START-TIME-~s-" fn))))
       (eval `(prog1 (defparameter ,v
                       ;;   Must use defparameter here.
                       -1)
                (declaim (fixnum ,v))))))
    (tablename
     (eval
      `(defglobal ,(hons-gentemp (suffix "MEMOIZE-HT-FOR-~s-" fn))
         nil)))
    (ponstablename
     (eval `(defglobal ,(hons-gentemp (suffix "PONS-HT-FOR-~s-" fn))
              nil)))
    (memoized-fn (eval `(defglobal ,(gensym "MEMOIZED-DEF") nil)))
    (top-level-call-p (gensym "TOP-LEVEL-CALL-P"))
    (start-hons-count (gensym "START-HONS-COUNT"))
    (start-free-bytes (gensym "START-FREE-BYTES"))
    (caller-callee    (gensym "CALLER-CALLEE"))
    (body-fn          (gensym "BODY"))
    (ans              (gensym "ANS"))
    (ans-p            (gensym "ANS-P"))
    (ma               (gensym "MA"))
    (caller           (gensym "CALLER"))
    (sts (loop for x in (union stobjs-in stobjs-out) when x
               collect (st-lst x)))
    (args          (gensym "ARGS"))
    argq defun success)
   (setq argq (if (cdr formals) args (car formals)))
   (setq defun
    `(defun ,fn ,formals ,@dcls (declare (ignorable ,@formals))
      (flet ((,body-fn () ,body))
       (let* ((,top-level-call-p (eql -1 ,start-time))
              (,start-time       (if ,top-level-call-p
                                     (internal-real-time)
                                   ,start-time))
              (,start-free-bytes (if (and ,top-level-call-p
                                          *report-bytes-used*)
                                     (free-bytes)
                                   -1))
              (,start-hons-count (if (and ,top-level-call-p
                                          *report-hons-calls*)
                                     *hons-call-counter*
                                   -1))
              (,caller-callee (the fixnum (+ *caller* ,fnn)))
              (,ma *memoize-array*)
              (,caller (the fixnum (* *max-memoize-fns* ,fnn)))
              (*caller* ,caller) ,ans ,ans-p ,args)
         (declare (ignorable ,ans ,ans-p ,args)
                  (fixnum ,caller
                          ,caller-callee
                          ,start-hons-count
                          ,start-free-bytes)
                  (type (simple-array fixnum (*)) ,ma))
         (safe-incf (aref ,ma ,caller-callee))
         (multiple-value-prog1
          (cond
           ((not ,condition-body) (,body-fn))
           (t
            ,@(if trace `((oftr "~%(~s passed arguments:  (" ',fn)
                          ,@(loop for v in formals append
                                  `((oftr "~& ~s = " ',v)
                                    (prine ,v *trace-output*)))
                          (oftr "~&)")))
            (when (null ,tablename)
              (safe-incf (aref ,ma (the fixnum (+ ,caller 2))))
              (setq ,tablename (mht))
              ,@(if (cdr formals) `((setq ,ponstablename (mht)))))
            ,@(if (cdr formals)
                  `((setq ,args (pist* ,ponstablename ,@formals))))
            (multiple-value-setq (,ans ,ans-p)
              (gethash ,argq ,tablename))
            (cond
             (,ans-p
              ,@(if trace `((oftr "~%Answer to ~s call remembered."
                                  ',fn)))
              (safe-incf (aref ,ma (the fixnum (1+ ,caller))))
              ,(cond ((eql 1 (length stobjs-out)) ans)
                     (t  (cons 'mv
                               (nconc
                                (loop for i below
                                      (1- (length stobjs-out))
                                      collect `(pop ,ans))
                                (list ans))))))
             (t ,(cond ((eql (length stobjs-out) 1)
                        `(setf (gethash ,argq ,tablename) (,body-fn)))
                       (t (let ((vars (loop for i
                                            below (length stobjs-out)
                                            collect (gensym "O"))))
                            `(mv-let ,vars (,body-fn)
                                     (setf (gethash ,argq ,tablename)
                                           (list* ,@vars))
                                     (mv ,@vars)))))))))
          (cond (,top-level-call-p
                 (let ((intr (internal-real-time)))
                   (declare (fixnum intr))
                   (when *report-hons-calls*
                     (safe-incf (aref ,ma
                                      (the fixnum
                                        (+ *ma-hons-calls-index*
                                           ,caller)))
                                (the fixnum
                                  (- *hons-call-counter*
                                     ,start-hons-count))))
                   (when (< *last-gc-exit-time* ,start-time)
                     (safe-incf
                      (aref ,ma
                            (the fixnum (+ *ma-bytes-used-index*
                                           ,caller)))
                      (the fixnum
                        (- ,start-free-bytes (free-bytes)))))
                   (when (> intr ,start-time)
                     (safe-incf (aref ,ma ,caller)
                                (the fixnum (- intr ,start-time)))))))
          ,@(if trace `((oftr "~%~s returned:~&" ',fn)
                        (prine (gethash ,argq ,tablename)
                               *trace-output*)
                        (oftr "~&)~%")
                        )))))))
   (when *memoize-debug*
     (our-syntax
      (let ((*print-pretty* t))
        (print defun *debug-io*)
        (force-output *debug-io*))))
   (unwind-protect
       (progn
         (loop for i fixnum from (* fnn *max-memoize-fns*)
               to (+ (* fnn *max-memoize-fns*) *max-symbol-to-fixnum*)
               do (setf (aref *memoize-array* i) 0))
         (setf (gethash fn *memoize-info-ht*)
               (make memoize-info-ht-entry
                     :tablename     tablename
                     :ponstablename ponstablename
                     :old-fn        old-fn
                     :memoized-fn   memoized-fn
                     :condition     condition
                     :inline        inline
                     :num           fnn
                     :sts           sts
                     :trace         trace
                     :start-time    start-time))
         (setf (gethash fnn *memoize-info-ht*) fn)
         (memoize-eval-compile defun)
         (setf (symbol-value memoized-fn) (symbol-function fn))
         (and condition
              (loop for s in sts do
                    (push tablename (symbol-value s))))
         (setq success t)
         fn)
     (unless success
       (setf (symbol-function fn) (symbol-value old-fn))
       (remhash fn *memoize-info-ht*)
       (remhash fnn *memoize-info-ht*)
       (loop for s in sts do
             (setf (symbol-value s)
                   (remove tablename (symbol-value s))))
       (ofe "~&; Memoize-fn:  Failed to memoize ~a." fn)))))))

(defun unmemoize-fn (fn)
  (eval `(maybe-untrace ,fn))
  (let* ((l (or (memoizedp-raw fn)
                (ofe "~&; Unmemoize-fn: ~a is not memoized." fn)))
         (tablename (access memoize-info-ht-entry l :tablename))
         (ponstablename (access memoize-info-ht-entry l
                                :ponstablename))
         (old-fn (access memoize-info-ht-entry l :old-fn))
         (memoized-fn (access memoize-info-ht-entry l :memoized-fn))
         (num (access memoize-info-ht-entry l :num))
         (sts (access memoize-info-ht-entry l :sts)))

; Note: condition is a first-class ACL2 function, not to be messed
; with here.

    (remhash fn *memoize-info-ht*)
    (setf (symbol-function fn) (symbol-value old-fn))
    (setf (symbol-value tablename) nil)
    (setf (symbol-value ponstablename) nil)
    (setf (symbol-value old-fn) nil)
    (setf (symbol-value memoized-fn) nil)
    (remhash num *memoize-info-ht*)
    (loop for s in sts do
          (setf (symbol-value s)
                (remove tablename (symbol-value s))))
    (when (eql 1 (hash-table-count *memoize-info-ht*))
      (memoize-init)))
  fn)

(defun maybe-unmemoize (fn)
  (when (memoizedp-raw fn) (unmemoize-fn fn)))

(defun memoized-functions ()
  (let (l)
    (maphash (lambda (fn v) (declare (ignore v))
               (when (symbolp fn) (push fn l)))
             *memoize-info-ht*)
    l))

(defun unmemoize-all ()

; A warning to would-be code improvers.  It would be a bad idea to
; redefine UNMEMOIZE-ALL to maphash over *memoize-info-ht* because of
; the ANSI rules concerning which hash table entries may be modified
; during a maphash.

  (loop for x in (memoized-functions) do (unmemoize-fn x)))

(defun memoize-unmemoized-starting-at
  (n &key (condition t) (inline t) (trace nil))
  (when (symbolp n) (setq n (event-number n)))
  (check-type n integer)
  (let* ((state *the-live-state*)
         (packs (get-global 'packages-created-by-defpkg state))
         (w (w state)))
    (labels
        ((no-use (pn fn)
           (let ((x (getprop fn pn t 'current-acl2-world w)))
             (or (eq x t) (member 'state x))))
         (unmemoizable (fn)
                       (or (memoizedp-raw fn)
                           (not (fboundp fn))
                           (eq (symbol-package fn)
                               *main-lisp-package*)
                           (gethash fn *memoize-reject-ht*)
                           (macro-function fn)
                           (special-form-or-op-p fn)
                           (no-use 'stobjs-in fn)
                           (no-use 'stobjs-out fn)
                           (no-use 'formals fn)
                           (and inline
                                (null (cltl-def-from-name
                                       fn nil w))))))
      (loop for p in (cons "ACL2" packs) do
            (do-symbols (fn p)
              (unless (or (unmemoizable fn) (< (event-number fn) n))
                (memoize-fn fn
                            :condition condition
                            :inline inline
                            :trace trace)))))))

(defun compliant-and-ideal ()
  (let* ((logic-fns
          (eval '(let ((world (w *the-live-state*)))
                   (strip-cadrs (set-difference-theories
                                 (function-theory :here)
                                 (universal-theory 'ground-zero))))))
         (ideal-fns (collect-non-common-lisp-compliants
                     logic-fns (w *the-live-state*))))
    (mv (set-difference-eq logic-fns ideal-fns) ideal-fns)))

(defun memoize-compliant-after
  (n &key (condition t) (inline t) (trace nil))
  (when (symbolp n) (setq n (event-number n)))
  (check-type n integer)
  (mv-let (c i)
          (compliant-and-ideal)
          (declare (ignore i))
          (loop for fn in c do
                (cond ((and (not (memoizedp-raw fn))
                            (> (event-number fn) n))
                       (memoize-fn fn
                                   :condition condition
                                   :inline inline
                                   :trace trace))))))

; Statistics Gathering Routines

(defun hons-statistics ()
  (our-syntax
   (oft "~&; Examining *hons-cdr-ht*:")
   (maphash (lambda (key value)
              (oft "~&; To ~s has been honsed:~%" key)
              (cond ((hash-table-p value)
                     (maphash (lambda (key v2) (declare (ignore v2))
                                (oft "~s, " key))
                              value))
                    (t (loop for pair in value do
                             (oft "~s, " (car pair))))))
            *hons-cdr-ht*)
   (oft "~&; End of *hons-cdr-ht* examination. ~
         ~%~%; Examining *hons-cdr-ht-eql*:")
   (maphash (lambda (key value)
              (oft "~%; To ~s has been honsed:~%" key)
              (cond ((hash-table-p value)
                     (maphash (lambda (key v2) (declare (ignore v2))
                                (oft "~s, " key))
                              value))
                    (t (loop for pair in value do
                             (oft "~s, " (car pair))))))
            *hons-cdr-ht-eql*)
   (oft "~%; End of *hons-cdr-ht-eql* examination. ~
         ~%~%; Examining *nil-ht*:~%")
   (oft "; To NIL has been honsed:~%")
   (maphash (lambda (key v2) (declare (ignore v2))
              (oft "~s, " key))
            *nil-ht*)
   (oft "~% End of *nil-ht* examination.")))

(defun print-alist (alist separation)
  (check-type separation (integer 0))
  (setq alist
        (loop for x in alist collect
              (progn
                (check-type x
                            (cons string
                                  (cons (or string (integer 0))
                                        null)))
                (list (car x)
                      (if (integerp (cadr x))
                          (numstr (cadr x))
                        (cadr x))))))
  (let* ((max1 (loop for x in alist maximize (length (car x))))
         (max2 (loop for x in alist maximize (length (cadr x))))
         (width (max *print-right-margin* (+ separation max1 max2))))
    (loop for x in alist do
          (fresh-line)
          (write (car x)
                 :case :downcase
                 :escape nil
                 :readably nil)
          (loop for i fixnum
                below (the fixnum
                        (- width (the fixnum (+ (length (car x))
                                                (length (cadr x))))))
                do (write-char #\Space))
          (princ (cadr x))))
  nil)

(defun hons-count ()
  (let ((n 0))
    (declare (integer n))
    (loop for tab in '(*hons-cdr-ht* *hons-cdr-ht-eql*) do
          (maphash (lambda (k v) (declare (ignore k))
                     (cond ((hash-table-p v)
                            (incf n (hash-table-count v)))
                           (t (incf n (length v)))))
                   (symbol-value tab)))
    (+ n (hash-table-count *nil-ht*))))

(defun hons-summary ()
  (-----)
  (our-syntax
   (let ((sssub 0) (nhonses 0) (nsubs 0))
     (declare (integer sssub nhonses nsubs))
     (loop for tab in '(*hons-cdr-ht* *hons-cdr-ht-eql*) do
           (maphash
            (lambda (k v) (declare (ignore k))
              (cond
               ((hash-table-p v)
                (incf sssub (hash-table-size v))
                (incf nhonses (hash-table-count v))
                (incf nsubs))
               (t (incf nhonses (length v)))))
            (symbol-value tab)))
     (incf nhonses (hash-table-count *nil-ht*))
     (print-alist
      `(,@(if (> *hons-call-counter* 0)
              `(("Hons hits/calls"
                 ,(let* ((c *hons-call-counter*)
                         (d (- c *hons-misses-counter*)))
                    (ofn "~,1e / ~,1e = ~,2f" d c (float (/ d c)))))))
        ,@(loop for tab in '(*hons-cdr-ht* *hons-cdr-ht-eql* *nil-ht*
                                           *hons-str-ht*)
                unless (eql 0 (hash-table-size (symbol-value tab)))
                collect
                (let* ((tabv (symbol-value tab))
                       (c (hash-table-count tabv))
                       (s (hash-table-size tabv)))
                  (list (ofn "~a count/size"
                             (string-downcase (symbol-name tab)))
                        (ofn "~,1e / ~,1e = ~,2f"
                             (numstr c) (numstr s) (float (/ c s))))))
        ("Number of sub tables" ,nsubs)
        ("Sum of sub table sizes" ,sssub)
        ("Number of honses" ,nhonses))
      5)))
  nil)

(defun memoize-statistics (&optional (fn (memoized-functions)))
  (our-syntax
   (cond ((listp fn)
          (mapc #'memoize-statistics fn))
         ((not (memoizedp-raw fn))
          (oft "~%; Memoize-statistics:  ~a is not memoized." fn))
         (t (let ((tb (symbol-value
                       (access memoize-info-ht-entry
                               (gethash fn *memoize-info-ht*)
                               :tablename))))
              (cond ((and tb (not (eql 0 (hash-table-count tb))))
                     (oft "~%; Memoized values for ~a." fn)
                     (maphash (lambda (key v)
                                (format t "~%~s~%=>~%~s" key v))
                              tb)))))))
  nil)

(defun number-of-calls (fn)
  (let ((fnn (the fixnum (symbol-to-fixnum fn))))
    (declare (fixnum fnn))
    (loop for i fixnum
          to (* *max-symbol-to-fixnum* *max-memoize-fns*)
          by *max-memoize-fns*
          sum (aref *memoize-array* (the fixnum (+ i fnn))))))

(defun number-of-hits (fn)
  (aref *memoize-array*
        (the fixnum (+ *ma-hits-index*
                       (the fixnum
                         (* *max-memoize-fns*
                            (the fixnum (symbol-to-fixnum fn))))))))

(defun number-of-mht-calls (fn)
  (aref *memoize-array*
        (the fixnum (+ *ma-mht-index*
                       (the fixnum
                         (* *max-memoize-fns*
                            (the fixnum (symbol-to-fixnum fn))))))))

(defun time-for-non-hits/call (fn)
  (let ((n (- (number-of-calls fn) (number-of-hits fn))))
    (if (zerop n) 0 (/ (total-time fn) n))))

(defun time/call (fn)
  (let ((n (number-of-calls fn)))
    (if (zerop n) 0 (/ (total-time fn) n))))

(defun hits/calls (fn)
  (let ((n (number-of-calls fn)))
    (if (zerop n) 0 (/ (number-of-hits fn) (float n)))))

(defun event-number (fn)
  (fgetprop fn 'absolute-event-number t (w *the-live-state*)))

; MEMOIZE-SUMMARY

(defparameter *memoize-summary-order* 'total-time)

(defglobal *memoize-summary-show-tables* t)

(defun hons-acons-summary ()
  (let ((count 0) (size 0) (number 0) last-key alist)
    (maphash
     (lambda (key v)
       (cond ((typep v 'fixnum)
              (push (list
                     (if (setq last-key (cdr (last key)))
                         last-key number)
                     (len key)
                     (len key))
                    alist)
              (incf number)
              (incf size (len key))
              (incf count (len key)))
             (t (push (list
                       (if (setq last-key (cdr (last key)))
                           last-key number)
                       (hash-table-size v)
                       (hash-table-count v))
                      alist)
                (incf number)
                (incf size (hash-table-size v))
                (incf count (hash-table-count v)))))
     *hons-acons-ht*)
    (cond (alist
           (oft "~&Hons-acons statistics")
           (print-alist
            (list (list "Count/size"
                        (ofn "~,1e/~,1e = ~,2f"
                             (numstr (hash-table-count
                                   *hons-acons-ht*))
                             (numstr (hash-table-size
                                   *hons-acons-ht*))
                             (/ (hash-table-count *hons-acons-ht*)
                                (hash-table-size
                                 *hons-acons-ht*))))
                  (list "Total of counts" count)
                  (list "Total of sizes" size))
            5)
           (oft "~&For each HONS-ACONS entry~%(name size count)")
           (loop for x in alist do (print x))))))

(defun memoize-summary-sort (fns)
  (let ((pairs
         (loop for fn in fns
               when
               (or (< 0 (total-time fn))
                   (not (eql -1 (execution-order fn))))
               collect (cons fn (funcall *memoize-summary-order*
                                         fn)))))
    (sort pairs #'> :key #'cdr)))

(defun execution-order (fn)
  (the fixnum (symbol-value
               (the symbol
                 (access memoize-info-ht-entry
                         (gethash fn *memoize-info-ht*)
                         :start-time)))))

(defn print-call-stack ()
  (let ((l nil) (x 0))
    (maphash (lambda (k v)
               (cond ((symbolp k)
                      (setq x (symbol-value
                               (the symbol
                                 (access memoize-info-ht-entry
                                         v
                                         :start-time))))
                      (when (not (eql -1 x))
                        (push (cons (string-downcase (symbol-name k))
                                    x)
                              l)))))
             *memoize-info-ht*)
    (setq l (sort l #'< :key #'cdr))
    (when (null (cdr l))
      (return-from print-call-stack))
    (oft "~&Time since the most recent outermost call.")
    (setq x (cdr (car (last l))))
    (setq l (loop for pair in l collect
                  (list (car pair)
                        (ofn "~,1e"
                             (/ (- x (cdr pair))
                                *float-ticks/second*)))))
    (print-alist l 5)))

(defun hons-calls (fn)
  (aref *memoize-array*
        (the fixnum (+ *ma-hons-calls-index*
                       (the fixnum
                         (* *max-memoize-fns*
                            (the fixnum (symbol-to-fixnum fn))))))))

(defun bytes-used (fn)
  (aref *memoize-array*
        (the fixnum (+ *ma-bytes-used-index*
                       (the fixnum
                         (* *max-memoize-fns*
                            (the fixnum (symbol-to-fixnum fn))))))))

(defun memoize-summary ()
 (when *in-memoize* (return-from memoize-summary nil))
 (let ((fn-pairs (memoize-summary-sort (memoized-functions))))
   (when fn-pairs
     (oft "~&(length (memoized-functions)) = ~a.  Order = ~a."
          (/ (1- (hash-table-count *memoize-info-ht*)) 2)
          *memoize-summary-order*))
   (when (null fn-pairs)
     (return-from memoize-summary nil))
   (loop for pair in fn-pairs do
        (let* ((fn (car pair))
               (l (gethash fn *memoize-info-ht*))
               (tablename
                (symbol-value (access memoize-info-ht-entry l
                                      :tablename)))
               (ponstablename (symbol-value
                               (access memoize-info-ht-entry
                                       l :ponstablename)))
               (start-time
                (symbol-value (access memoize-info-ht-entry
                                      l :start-time)))
               (num (access memoize-info-ht-entry
                            l :num))
               (nhits (number-of-hits fn))
               (nmht (number-of-mht-calls fn))
               (ncalls (number-of-calls fn))
               (hons-calls (hons-calls fn))
               (bytes-used (bytes-used fn)))
          (print-alist
           `((,(let ((in-progress-str
                      (if (eql start-time -1) "" " (running)")))
                 (if (eql 0 nhits)
                     (ofn "~a calls=~a"
                          (string-upcase
                           (symbol-name fn))
                   in-progress-str)
                   (ofn "~a hits/calls=~a"
                        (string-upcase (symbol-name fn))
                        in-progress-str)))
              ,(if (eql 0 nhits)
                   (ofn "~a" (numstr ncalls))
                 (ofn "~,1e / ~,1e = ~,1f"
                      (numstr nhits)
                      (numstr ncalls)
                      (/ nhits (float ncalls)))))
             ,@(if (>= nmht 2) `(("Calls to mht" ,(numstr nmht))))
             ("time of all outermost calls"
              ,(numstr (total-time fn)))
             ("time per outermost call"
              ,(numstr (time/call fn)))
             ,@(if (and (> bytes-used 0) *report-bytes-used*)
                   `(("bytes-used" ,(numstr bytes-used))))
             ,@(if (and (> hons-calls 0) *report-hons-calls*)
                   `(("hons-calls" ,(numstr hons-calls))))
             ,@(if (not (eql 0 nhits))
                   `(("time per missed outermost call"
                      ,(ofn "~,1e"
                            (time-for-non-hits/call fn)))))
             ,@(loop
                for i fixnum to *max-symbol-to-fixnum*
                when (gethash i *memoize-info-ht*)
                nconc
                (let ((calls
                       (aref *memoize-array*
                             (the fixnum
                               (+ num
                                  (the fixnum
                                    (* i
                                       *max-memoize-fns*
                                       )))))))
                  `(,@(if (> calls 0)
                          `((,(ofn "calls from ~a"
                                   (if (eql i 0)
                                       "unmemoized functions"
                                     (fixnum-to-symbol i)))
                             ,calls))))))
             .
             ,(if (not *memoize-summary-show-tables*) nil
                (let ((spsub 0) (nponses 0) (npsubs 0))
                  (and
                   ponstablename
                   (maphash
                    (lambda (key value)
                      (declare (ignore key))
                      (cond
                       ((hash-table-p value)
                        (incf spsub (hash-table-size value))
                        (incf nponses
                              (hash-table-count value))
                        (incf npsubs))
                       (t (incf nponses (length value)))))
                    ponstablename))
                  `(,@(and
                       tablename
                       `((,(ofn "memoize table count/size")
                          ,(ofn "~,1e / ~,1e = ~,2f"
                                (numstr (hash-table-count tablename))
                                (numstr (hash-table-size tablename))
                                (/ (hash-table-count tablename)
                                   (+ .001 (hash-table-size
                                            tablename)))))))
                      ,@(and
                         ponstablename
                         `(("pons table count/size"
                            ,(ofn "~,1e / ~,1e = ~,2f"
                                  (numstr (hash-table-count
                                        ponstablename))
                                  (numstr (hash-table-size
                                        ponstablename))
                                  (/ (hash-table-count
                                      ponstablename)
                                     (+ .001 (hash-table-size
                                              ponstablename)))))
                           ("number of pons sub tables"
                            ,npsubs)
                           ("sum of sub pons table sizes"
                            ,spsub)
                           ("sum of sub pons table counts"
                            ,nponses)))))))
           5)))))


; State refreshment routines for hons and memoize

(defun init-hons-acons-table ()
  (setq *hons-acons-ht* (mht :test #'eq :weak :key)))

; The ACL2 persistent-hons-table, which is updated by defhonst, should
; be an alist whose keys are honses to be preserved as honses through
; any cleaning of hash tables, but not through an init-hash-tables.

(defglobal *init-hash-tables-done* nil)
(defun init-hash-tables ()
  ; Can be repeatedly requested by the user.
  (setq *hons-cdr-ht* (mht :size *hons-cdr-ht-size*
                           :weak :key :test #'eq))
  (setq *hons-cdr-ht-eql* (mht))
  (setq *nil-ht* (mht :size *nil-ht-size* :weak :value))
  (init-hons-acons-table)
  (setq *hons-str-ht* (mht :test #'equal :weak :value))
  (setq *hons-copy-aux-ht* (mht :test #'eq))
  (setq *init-hash-tables-done* t))

(defglobal *rehons-culprit* nil)

(defun rehons (x)
  (cond ((and (consp x)
              (not (eq x (hons-copy1-consume x))))
         (setq *rehons-culprit* x)
         (ofe "~%; clear-hash-tables:  Error: ** failed to ~
                         rehons *rehons-culprit*."))
        (t (maybe-str-hash x))))

(defun clear-hash-tables ()

; Should this function save any honses from *hons-acons-ht*?
; Currently it saves not only persistent honses, but also honses that
; are keys of, or keys of keys of, the *hons-acons-ht*.  An optional
; argument could presumably be added to make these additional saves
; optional (but then what should the default be for this optional
; argument?).

; hons-let assumes that clrhash is not used here.

  (let (l)
    (declare (dynamic-extent l))
    (maphash (lambda (k v) (declare (ignore v))
               (if (honsp k) (push k l)))
             *hons-acons-ht*)
    (setq *hons-cdr-ht* (mht ; :size *hons-cdr-ht-size*
                         :test #'eq
                         :weak :key))
    (setq *hons-cdr-ht-eql* (mht))
    (setq *nil-ht* (mht ; :size *nil-ht-size*
                    :weak :value))
    (loop for x in (table-alist 'persistent-hons-table
                                (w *the-live-state*))
          do (when (car x) (rehons (car x))))
    (mapc #'rehons l)
    (maphash (lambda (k v)
               (cond ((and (consp k) (honsp k))
                      ;; all parts of k are already honsed.
                      nil)
                     ((hash-table-p v)
                      (maphash (lambda (k v)
                                 (declare (ignore v))
                                 (rehons k))
                               v))
                     (t (loop for pair in k do
                              (rehons (car pair))))))
             *hons-acons-ht*))
  nil)

(defun empty-ht-p (x)
  (and (hash-table-p x)
       (eql 0 (hash-table-count x))))

; *memoized-reject-ht* maps symbols that should not be memoized to t.
; Some of these symbols are taboo because they are used in the
; memoization machinery.  Others are used in the watch machinery.
; Others are read-eval-loop type functions that are strange to
; memoize, since they are likely to do whatever they are asked to do.

(defglobal *car-cdr-list*
  '(car cdr caar cadr cdar cddr caaar
        cadar cdaar cddar caadr caddr
        cdadr cdddr caaaar cadaar cdaaar
        cddaar caadar caddar cdadar cdddar
        caaadr cadadr cdaadr cddadr caaddr
        cadddr cdaddr cddddr))

(defglobal *memoize-reject-ht*
  (let ((h (mht :test #'eq)))
    (loop for fn in '(
                      #+rdtsc
                      ccl::rdtsc64
                      access-event-tuple-number
                      assoc-eq
                      assoc-no-error-at-end
                      backquote
                      backquote-lst
                      clear-memo-tables
                      cltl-def-from-name
                      compact-read-init
                      current-form
                      current-values-of-some-vars
                      dcls
                      defconst-fn
                      ev
                      ev-fncall
                      ev-fncall-rec
                      ev-for-trans-eval
                      ev-rec
                      event-number
                      fetch-from-zap-table
                      fgetprop
                      fixnum-to-symbol
                      fmt0
                      fmt1
                      numstr
                      gctime
                      get-doc-string
                      get-global
                      getpid$
                      getprop-default
                      global-symbol
                      global-val
                      hits/calls
                      hons-acons-summary
                      hons-calls/sec-run-time
                      hons-copy
                      hons-copy1
                      hons-copy1-consume
                      hons-copy3
                      hons-copy3-consume
                      hons-gentemp
                      hons-hits/calls
                      hons-init
                      hons-summary
                      init-hash-tables
                      intern-in-package-of-symbol
                      ld-fn
                      ld-fn-body
                      ld-loop
                      ld-read-command
                      ld-read-eval-print
                      lookup-world-index
                      lookup-world-index1
                      max-absolute-event-number
                      maybe-str-hash
                      member-equal
                      memoize-array-grow
                      memoize-compliant-after
                      memoize-eval-compile
                      memoize-fn
                      memoize-init
                      memoize-statistics
                      memoize-summary
                      memoize-summary-sort
                      memoize-unmemoized-after
                      memoized-functions
                      memoizedp-raw
                      newline
                      number-of-args
                      number-of-calls
                      #+openmcl number-of-cpus-on-this-machine
                      number-of-gcs
                      number-of-hits
                      number-of-mht-calls
                      number-of-mht-calls
                      number-of-return-values
                      ofe
                      open-output-channel
                      pons
                      pons-calls/sec-run-time
                      pons-hits/calls
                      princ$
                      print-alist
                      print-rules-summary
                      print-summary
                      print-time-summary
                      profile
                      profile-unmemoized
                      protected-eval-with-proofs
                      prove
                      prove-loop
                      prove-loop1
                      raw-ev-fncall
                      read-object
                      read-standard-oi
                      record-accessor-function-name
                      run-time-since-watch-start
                      safe-incf-error
                      scan-to-landmark-number
                      set-w
                      sgetprop
                      sh
                      simple-translate-and-eval
                      st-lst
                      suffix
                      symbol-to-fixnum
                      symbol-to-fixnum-create
                      table-alist
                      thm-fn
                      time-for-non-hits/call
                      time-of-last-watch-update
                      time-since-watch-start
                      time-since-watch-start
                      time/call
                      timer-error
                      too-long
                      total-time
                      trans-eval
                      translate1
                      translate11
                      translate11-mv-let
                      unmemoize-all
                      unmemoize-fn
                      w
                      watch
                      watch-interrupt
                      watch-function
                      watch-kill
                      watch-run-time
                      watch-time
                      waterfall
                      waterfall-print-clause
                      waterfall-step
                      waterfall0
                      waterfall1
                      waterfall1-lst

                      1+ + * - <= < > >= =
                      aref arrayp atom apply
                      car cdr cons consp
                      eq eql error
                      format
                      get-internal-real-time
                      get-internal-run-time
                      gethash
                      hash-table-p
                      list
                      make-hash-table
                      max
                      not null
                      prin1 princ print print-object
                      stringp svref symbolp
                      typep
                      write-char write-byte write)
          do (setf (gethash fn h) t))
    (loop for fn in (cddr *car-cdr-list*) do
          (setf (gethash (ofni "safe-~a" fn) h) t)
          (setf (gethash (ofni "gentle-~a" fn) h) t))
    h))

(defglobal *memoize-init-done* nil)
(defun memoize-init ()
  ; Should only be repeated by unmemoize-fn.
  (setq *memoize-info-ht* (mht))
  (setf (gethash 0 *memoize-info-ht*) "outside-caller")
  (setq *max-symbol-to-fixnum* 2)
  (setq *max-memoize-fns* 30)
  (setq *memoize-array*
        (make-array (* *max-memoize-fns* *max-memoize-fns*)
                    :element-type 'fixnum
                    :initial-element 0))
  (setq *memoize-init-done* t))

(defun clear-memoize-table (k)
  (let (l)
    (when (and (symbolp k)
               (setq l (gethash k *memoize-info-ht*)))
      (unless (empty-ht-p
               (symbol-value
                (the symbol (access memoize-info-ht-entry
                                    l :tablename))))
        (setf (symbol-value
               (the symbol
                 (access memoize-info-ht-entry l :tablename)))
              nil))
      (unless (empty-ht-p
               (symbol-value
                (the symbol (access memoize-info-ht-entry
                                    l :ponstablename))))
        (setf (symbol-value
               (the symbol
                 (access memoize-info-ht-entry l :ponstablename)))
              nil))))
  nil)

(defun clear-memoize-tables ()
  "Clears away all memoized answers."

; May be called repeatedly by the user.

  (mapc #'clear-memoize-table (memoized-functions))
  nil)

(defun clear-counters ()

  "Clears counting and timing information for all memoized functions."

  (setq *pons-call-counter*   0)
  (setq *pons-hit-counter*    0)
  (setq *hons-call-counter*   0)
  (setq *hons-misses-counter* 0)

  (setq *memoize-array*
        (make-array (* *max-memoize-fns* *max-memoize-fns*)
                    :element-type 'fixnum
                    :initial-element 0)))

(defun clear-hash-and-memoize-tables ()
  (clear-hash-tables)
  (clear-memoize-tables))

(defun flush-hons-get-hash-table-link (x)

; FLUSH-HONS-GET-HASH-TABLE-LINK breaks the link between X and a hash
; table in *HONS-ACONS-HT*, if such a link exists, thus probably
; permitting the garbage collection of that table.

  (unless (atom x) (remhash x *hons-acons-ht*))
  x)


; Hash-consing while reading: HONS-READ-OBJECT and HONS-READ.

; Hash consing when reading is implemented via a change to the
; readtable for the character including open parenthesis, close
; parenthesis, and the consing dot.

; *** NOTE:  The following collection of functions is just that: a
;            collection.  Unless you understand everything about the
;            various read table macros, then don't touch this code!

; See matching comment below.

; Note: our implementation of the #=/## reader, which we built because
; some Lisps would not us get past #500 or so, does not comply with
; ANSI at least in this regard: it does not allow the entry of looping
; structures as in '(#0= (3 #0#)), which no problem for ACL2 uses.

(defglobal *hons-readtable* (copy-readtable *acl2-readtable*))
(declaim (readtable *acl2-readtable* *hons-readtable*))

; WARNING: Any call of read using *hons-readtable* as *readtable*
; needs to worry about the possible top-level return of
; *close-paren-obj* and *dot-obj*, which are simply part of the
; fiction whereby we read while honsing.  Those two objects should
; absolutely not be returned as the value of an ACL2 function.  See,
; for example, the definition of HONS-READ.

(defglobal *close-paren-obj* '(#\)))
(defglobal *dot-obj*         '(#\.))

(defun nonsense (x)
  (or (eq x *close-paren-obj*) (eq x *dot-obj*)))

(defun check-nonsense (x)
  (cond ((nonsense x)
         (hread-error "~&;  Illegal object: ~a." (car x)))))

(defun hread-error (string &rest r)
  (our-syntax
   (let* ((stream *standard-input*)
          (*standard-output* *error-output*)
          (*standard-input* *debug-io*))
     (apply #'format *error-output* string r)
     (cond ((and (streamp stream) (file-position stream))
            (format *error-output*
                    "~&; near file-position ~a in stream ~a."
                    (file-position stream) stream)))
     (ofe "hread"))))

(defun illegal-error1 (x)
  (hread-error "~&; ** Error:  Illegal:  ~s." x))

(defun illegal-error2 (stream char)
  (declare (ignore stream))
  (illegal-error1 char))

(defun close-paren-read-macro (stream char)
  (declare (ignore char stream))
  (if *read-suppress* (illegal-error1 #\)))
  *close-paren-obj*)

(defun dot-read-macro (stream char)
  (declare (ignore char))
  (if *read-suppress* (illegal-error1 #\.))
  (let ((ch (peek-char nil stream t nil t)))
    (cond ((or (member ch '(#\( #\) #\' #\` #\, #\" #\;
                            #\Tab #\Space #\Newline))
               (eql 13 (char-code ch))
               (multiple-value-bind (fn nonterminating)
                   (get-macro-character ch)
                 (and fn (not nonterminating))))
           *dot-obj*)
          (t (let ((*readtable* *acl2-readtable*))
               (unread-char #\. stream)
               (read nil t nil t))))))

(defun hons-read-list ()
  (let ((o (read nil t nil t)))
    (cond
     ((eq o *close-paren-obj*) nil)
     ((eq o *dot-obj*)
      (let ((lo (read nil t nil t))
            (lp (read nil t nil t)))
        (check-nonsense lo)
        (cond
         ((eq lp *close-paren-obj*) lo)
         (t (illegal-error1 #\.)))))
     (t (hons-normed (maybe-str-hash o) (hons-read-list))))))

(defun hons-read-list-top ()
  (let ((o (read nil t nil t)))
    (cond
     ((eq o *close-paren-obj*) nil)
     (t (check-nonsense o)
        (hons-normed (maybe-str-hash o)
                     (hons-read-list))))))

(defun hons-read-reader (stream char)
  (declare (ignore char))
  (cond (*read-suppress*
         (unread-char #\( stream)
         (let ((*readtable* *acl2-readtable*))
           (read nil t nil t)))
        (t (hons-read-list-top))))

(defun hons-read (&optional i (eep t) eofv rp)
  (let* ((*readtable* *hons-readtable*)
         (*standard-input* (or i *standard-input*)))
    (let ((x (read nil eep eofv rp)))
      (check-nonsense x)
      x)))

(defun hons-read-object (channel state-state)
  (let ((*acl2-readtable* *hons-readtable*))
    (mv-let
     (eofp obj state)
     (read-object channel state-state)
     (check-nonsense obj)
     (mv eofp obj state))))

(defun hons-read-file (file-name)
  (with-open-file (file file-name)
    (let (temp ans (eof (cons nil nil)))
      (declare (dynamic-extent eof))
      (loop (setq temp (hons-read file nil eof nil))
            (if (eq eof temp)
                (return (hons-copy1-consume (nreverse ans))))
            (setq ans (cons temp ans))))))

;  Compact printing and reading procedures.

(defglobal *compact-print-file-ht* (mht))
(defglobal *compact-read-file-ht* (mht))
(defglobal *compact-print-file-n* 0)
(declaim (fixnum *compact-print-file-n*))
(defglobal *space-owed* nil)
(declaim (type boolean *space-owed*))

(defun compact-print-file-scan (x)
  (unless (or (and (symbolp x)
                   (let ((p (symbol-package x)))
                     (or (eq p *main-lisp-package*)
                         (eq p *package*)))
                   (<= (length (symbol-name x)) 4))
              (and (stringp x) (<= (length x) 2))
              (and (integerp x) (< -100 x 1000))
              (characterp x))
    (let ((g (gethash x *compact-print-file-ht*)))
      (unless (or (atom x) (eq g 'give-it-a-name))
        (compact-print-file-scan (car x))
        (compact-print-file-scan (cdr x)))
      (unless (eq g 'give-it-a-name)
        (setf (gethash x *compact-print-file-ht*)
              (if g 'give-it-a-name 'found-at-least-once))))))

(defmacro space-if-necessary ()
  '(when *space-owed* (write-char #\Space) (setq *space-owed* nil)))

(defun compact-print-file-help (x hash)
  (cond ((typep hash 'fixnum)
         (space-if-necessary)
         (write-char #\#)
         (princ hash)
         (write-char #\#))
        (t (cond ((eq hash 'give-it-a-name)
                  (let ((n *compact-print-file-n*))
                    (declare (fixnum n))
                    (when (eql n most-positive-fixnum)
                      (ofe "~&; *compact-print-file-n* overflow."))
                    (setq n (the fixnum (+ 1 n)))
                    (setq *compact-print-file-n* n)
                    (setf (gethash x *compact-print-file-ht*) n)
                    (space-if-necessary)
                    (write-char #\#)
                    (princ n)
                    (write-char #\=))))
           (cond
            ((atom x)
             (space-if-necessary)
             (prin1 x)
             (setq *space-owed* t))
            (t (write-char #\()
               (setq *space-owed* nil)
               (loop (compact-print-file-help
                      (car x)
                      (gethash (car x) *compact-print-file-ht*))
                     (cond
                      ((null (cdr x))
                       (write-char #\))
                       (setq *space-owed* nil)
                       (return))
                      ((or (progn
                             (setq hash
                                   (gethash (cdr x)
                                            *compact-print-file-ht*))
                             (or (eq hash 'give-it-a-name)
                                 (typep hash 'fixnum)))
                           (atom (cdr x)))
                       (space-if-necessary)
                       (write-char #\.)
                       (setq *space-owed* t)
                       (compact-print-file-help (cdr x) hash)
                       (write-char #\))
                       (setq *space-owed* nil)
                       (return))
                      (t (pop x)))))))))

(defun compact-print-file (data file-name)
  (setq *compact-print-file-ht* (mht))
  (setq *compact-print-file-n* 0)
  (setq *space-owed* nil)
  (with-open-file (*standard-output* file-name
                                     :direction :output
                                     :if-does-not-exist :create
                                     :if-exists :supersede)
    (our-syntax
     (compact-print-file-scan data)
     (compact-print-file-help
      data (gethash data *compact-print-file-ht*))
     (setq *compact-print-file-ht* (mht))))
  file-name)

(defun ns-=-reader (stream subchar arg)
  (declare (ignore stream subchar))
  (when (gethash arg *compact-read-file-ht*)
    (hread-error
     "~&; ns-=-reader:  ** Error:  #~a= is already defined to be ~s."
     arg (gethash arg *compact-read-file-ht*)))
  (setf (gethash arg *compact-read-file-ht*) (read nil t nil t)))

(defun ns-ns-reader (stream subchar arg)
  (declare (ignore stream subchar))
  (or (gethash arg *compact-read-file-ht*)
      (hread-error "~&; ns-ns-reader:  ** Error:  meaningless #~a#."
                   arg)))

(defglobal *compact-read-file-readtable*
  (copy-readtable *hons-readtable*))
(declaim (readtable *compact-read-file-readtable*))

(defun compact-read-file (fn)
  (setq *compact-read-file-ht* (mht))
  (with-open-file (*standard-input* fn)
    (our-syntax
     (let* ((*readtable* *compact-read-file-readtable*)
            (eof (cons nil nil))
            (p (read)))
       (check-nonsense p)
       (unless (eq (read nil nil eof) eof)
         (ofe "~%; compact-read-file: ** Error:  ~a has too many ~
                forms." fn))
       (setq *compact-read-file-ht* (mht))
       (hons-copy1-consume p)))))


; NEWICK PARSING

; A parser for Newick expressions conforming to
; http://evolution.genetics.washington.edu/phylip/newicktree.html and
; the last pages of:  "NEXUS: An Extensible File Format for Systematic
; Information," David R. Maddison; David L. Swofford; Wayne
; P. Maddison Systematic Biology, Vol. 46, No. 4. (Dec., 1997),
; pp. 590-621.

; Onto *NEWICK-INTERIOR-NODE-ALIST* we accumulate conses that pair
; names to interior nodes of the Newick expression currently being
; read.
(defglobal *newick-interior-node-alist* nil)

; *NEWICK-IGNORE-BRANCH-LENGTHS* determines whether a branch length is
; to be attached to Newick expression (by consing the expression onto
; the branch length), or simply ignored.
(defglobal *newick-ignore-branch-lengths* nil)

; *NEWICK-STR* an adjustable one-dimensional array of characters into
; which we accumulate the characters of a token as it is read.
(defglobal *newick-str*
  (make-array 60 :element-type 'character :adjustable t
              :fill-pointer 0))
(declaim (type (array character (*)) *newick-str*))

(defun newick-digit (c)

; NEWICK-DIGIT returns, for a character, the value of the character as
; a decimal digit, if it is a decimal digit, and otherwise returns
; NIL.

  (declare (character c))
  (case c (#\0 0) (#\1 1) (#\2 2)
        (#\3 3) (#\4 4) (#\5 5) (#\6 6)
        (#\7 7) (#\8 8) (#\9 9)))

(defmacro newick-ordinary-mac ()
  `(case c
     ,@(loop for c in (coerce "ABCDEFGHIJKLMNOPQRSTUVWXYZ" 'list)
             collect (list (list c (char-downcase c)) c))
     ,@(loop for c in (coerce "0123456789.<>~!@#$%^&|?" 'list)
             collect (list c c))
     (#\_ #\Space)))

(defun newick-ordinary (c)

; NEWICK-ORDINARY returns a non-NIL value on the characters that may
; occur in tokens outside of single quotes.  For a lowercase
; character, the corresponding uppercase character is returned.  For
; the underscore character, the space character is returned.
; Otherwise, the character itself is returned.

  (declare (character c))
  (newick-ordinary-mac))

(defmacro newick-white-mac ()
  `(case c
     ((#\Space #\Newline #\Tab
       ,(code-char 0) ,(code-char 1) ,(code-char 2)
       ,(code-char 3) ,(code-char 4) ,(code-char 5)
       ,(code-char 6)
       ,(code-char 12) ; CNTL-L
       ,(code-char 13) ; CNTL-M
       )
     t)))

(defun newick-white-p (c)

; NEWICK-WHITE-P returns non-NIL on characters that are white.

  (declare (character c))
  (newick-white-mac))

(defun newick-special-p (c)

; NEWICK-SPECIAL-P returns non-NIL on the five punctuation characters
; for Newick expressions.

  (case c ((#\, #\( #\) #\: #\;) t)))

(defun newick-illegal (x) (ofe "~&; Illegal:  ~a" x))

(defun newick-consume-comment ()

; When NEWICK-CONSUME-COMMENT is called, an open square bracket must
; have just been read.  NEWICK-CONSUME-COMMENT reads forward up to and
; including the balancing close square bracket.

  (let ((n 1))
    (declare (fixnum n))
    (loop (let ((ch (read-char nil t nil t)))
            (declare (character ch))
            (cond ((eql ch #\[) (safe-incf n))
                  ((eql ch #\])
                   (when (eql n 1) (return))
                   (setq n (the fixnum (1- n)))))))))

(defun newick-read-token ()

; If the next not-commented-out, non-white character is "special", we
; return it.  Otherwise, we return a symbol, which represents a token.

  (let ((p (read-char nil t nil t)) x)
    (declare (character p))
    (cond ((newick-white-p p) (newick-read-token))
          ((newick-special-p p) p)
          ((setq x (newick-ordinary p))
           (let ((s *newick-str*) (a *acl2-package*))
             (setf (fill-pointer s) 0)
             (vector-push (the character x) s)
             (loop
              (setq p (read-char nil t nil t))
              (cond ((setq x (newick-ordinary p))
                     (vector-push-extend (the character x) s))
                    ((eql p #\[) (newick-consume-comment))
                    (t (unread-char p) (return))))
             (or (find-symbol s a) (values (intern (copy-seq s) a)))))
          ((eql p #\[) (newick-consume-comment) (newick-read-token))
          ((eql p #\')
           (let ((c (read-char nil t nil t))
                 (flg nil) (s *newick-str*) (a *acl2-package*))
             (setf (fill-pointer s) 0)
             (loop
              (when (eql c #\')
                (cond ((eql #\' (peek-char nil nil t nil t))
                       (read-char nil t nil t))
                      (t (return))))
              (when (newick-ordinary c) (setq flg t))
              (vector-push-extend (char-upcase c) s)
              (setq c (read-char nil t nil t)))
             (unless flg
               (ofe "~&; Newick-read-token: ** Error: a Newick name ~
                     may not consist entirely of whitespace and ~
                     punctuation, ~%as does: ~s." s))
             (or (find-symbol s a) (values (intern (copy-seq s) a)))))
          (t (newick-illegal p)))))

(defun newick-read-branch-length ()
; P is the most recently read character.
; X is the digit value of P, if P is a digit; otherwise X is NIL.
; PAST-DOT, when non-NIL, is the number of digits we have read after a
;    decimal point and before any e or E.
; POS is T iff we have read an initial plus sign.
; NEG is T iff we have read an initial minus sign.
; ANS is the accumulated, nonnegative coefficient value.
; PAST-E is T iff we have read an e or E.
; ENEG is T iff we have read an e or E followed by a minus sign.
; EANS is the accumulated, nonnegative. exponent value.

; branch-lengths are represented in the following somewhat strange
; way.  An integer represents itself.  A Common Lisp complex number
; #c(i j) represents (* i (expt 10 j)).

  (let* ((ans 0) (eans 0) p x past-dot past-e pos neg eneg
         (ig *newick-ignore-branch-lengths*))
    (flet ((gp ()
               (loop (setq p (read-char nil t nil t))
                     (unless (eql p #\[) (return))
                     (newick-consume-comment))
               (setq x (newick-digit p))))
      ;; 1. Get past leading spaces, comments
      (gp)
      (loop (if (or x (not (newick-white-p p)))
                (return)
              (gp)))
      ;; 1.5 Get past .+-.
      (loop (cond (x (return))
                  ((eql p #\-)
                   (when (or pos neg past-dot) (newick-illegal p))
                   (setq neg t))
                  ((eql p #\.)
                   (when past-dot (newick-illegal p))
                   (setq past-dot 0))
                  ((eql p #\+)
                   (when (or pos neg past-dot) (newick-illegal p))
                   (setq pos t))
                  (t (newick-illegal p)))
            (gp))
      ;; 2. consume all remaining characters.
      (loop
       (cond (x (cond (ig)
                      (past-e (setq eans (+ (* eans 10) x)))
                      (t (setq ans (+ (* ans 10) x)))))
             ((or (eql p #\e) (eql p #\E))
              (if past-e (newick-illegal p))
              (setq past-e t)
              (let ((c (peek-char nil nil t nil t)))
                (cond ((eql c #\+)
                       (read-char nil t nil t)
                       (setq c (peek-char nil nil t nil t)))
                      ((eql c #\-)
                       (setq eneg t)
                       (read-char nil t nil t)
                       (setq c (peek-char nil nil t nil t))))
                (cond ((newick-digit c))
                      (t (newick-illegal c)))))
             ((eql p #\.)
              (if (or past-dot past-e) (newick-illegal p))
              (setq past-dot -1))
             (t (unread-char p)
                (if (or ig (eql ans 0))
                    (return-from newick-read-branch-length 0))
                (if neg (setq ans (- ans)))
                (if eneg (setq eans (- eans)))
                (if past-dot (decf eans past-dot))
                (loop 
                 (if (eql eans 0)
                     (return-from newick-read-branch-length ans))
                 (multiple-value-bind (q r) (floor ans 10)
                   (cond ((eql r 0)
                          (incf eans)
                          (setq ans q))
                         (t (return-from newick-read-branch-length
                              ;; next form could be replaced with
                              ;; (* ans (expt 10 eans))
                              (complex ans eans)))))))) 
       (if (and past-dot (not past-e) (not ig))
           (incf past-dot))
       (gp)))))

(defun newick-cleanup-help (p x)
  (cond ((eql p #\:)
         (let ((r (newick-read-branch-length)))
           (cond (*newick-ignore-branch-lengths* x)
                 (t (hons-normed x r)))))
        ((eql p #\() (newick-illegal p))
        ((characterp p)
         (unread-char p)
         x)
        (t (newick-illegal p))))

(defun newick-cleanup (x)

; NEWICK-CLEANUP returns the interior node (subtree), X, possibly
; consed to the node's branch length.  Before reading a colon,
; followed by branch length, a name for X may be read; if so, a pair
; will be added to *NEWICK-INTERIOR-NODE-ALIST* associating the name
; with X.  The final comma, semicolon, or close paren is left unread.

  (let ((p (newick-read-token)))
    (cond
     ((and (consp x) (symbolp p))
      (when (assoc p *newick-interior-node-alist*)
        (ofe "%&; Newick-cleanup: ** Error: An interior node ~
                name, such as ~s, may not be twice given a value." p))
      (push (cons p x) *newick-interior-node-alist*)
      (newick-cleanup-help (newick-read-token) x))
     (t (newick-cleanup-help p x)))))

(defun newick-read-list (flg)

; When NEWICK-READ-LIST is called, the open paren of a list has
; already been read.  We expect a comma, a close paren, or an element
; of the list.

; If FLG is NIL, an immediately encountered comma or close paren is to
; be read as a NIL, then as itself, helping us handle such Newick
; expressions such as (,(),(,),,).  If FLG is T, then an immediately
; encountered comma or close paren is simply to be skipped over.

; NOTE:  After calling NEWICK-READ-LIST, a call to NEWICK-CLEANUP
;        must be made.
  (let ((p (newick-read-token)))
    (cond
     ((null flg)
      (cond ((eql p #\)) (hons-normed nil nil))
            ((eql p #\,) (hons-normed nil (newick-read-list nil)))
            ((symbolp p) (hons-normed (newick-cleanup p)
                                      (newick-read-list t)))
            ((eql p #\()
             (hons-normed (newick-cleanup (newick-read-list nil))
                          (newick-read-list t)))
            (t (newick-illegal p))))
     ((eql p #\)) nil)
     ((eql p #\,)
      (setq p (newick-read-token))
      (cond ((symbolp p) (hons-normed (newick-cleanup p)
                                      (newick-read-list t)))
            ((eql p #\()
             (hons-normed (newick-cleanup (newick-read-list nil))
                          (newick-read-list t)))
            ((eql p #\,)
             (hons-normed nil (newick-read-list nil)))
            ((eql p #\)) (hons-normed nil nil))
            (t (newick-illegal p))))
      (t (newick-illegal p)))))

(defun newick-read-file-error-report (osi ch)

; When NEWICK-READ-FILE causes an error, we print its location, if
; feasible.

  (rotatef *standard-input* osi)
  (setq *newick-interior-node-alist* nil)
  (our-syntax
   (let* ((fp (file-position osi))
          (fl (and fp (file-length osi)))
          (p (and fp (max 0 (- fp 30))))
          (d (and fp (max 0 (- fp p))))
          (e *error-output*))
     (cond ((and fp (eql fp fl))
            (format e "~%; At end of file."))
           (fp (format e "~%; File position: ~a" fp)))
     (when (and fp (file-position osi p))
       (format e "~%; Last ~d chars read:" d)
       (loop for i below d do
             (write-char (read-char osi t nil t) e)))
     (if (characterp ch)
         (ofe "~%; Newick-read-file: ** Error.")
       (ofe "~&; Newick-read-file: Missing semicolon.")))))

; NEWICK-READ-FILE reads a file of Newick expressions.  Unless an
; error occurs, a proper list is returned each of whose elements is a
; cons or a string.  A string represents the contents of one line of
; the file encountered outside a Newick expression.  A cons has as its
; CAR a honsp+consp representing a Newick expression and as its CDR an
; alist pairing symbols with interior nodes of the expression.  If the
; first non-white, not-commented-out character on a line not within a
; Newick expression is not an open paren, the entire line is read as a
; string.

; The accumulation of strings depends upon the implementation of
; FILE-POSITION; if FILE-POSITION always returns NIL, strings are not
; accumulated at all.

(defun newick-read-file (file-name)
  (let ((osi *standard-input*) ch)
    (with-open-file (*standard-input* file-name)
      (let ((si *standard-input*) st ans)
        (loop (setq st (file-position si))

; Set CH to next non-white, not-commented-out character.

              (loop
               (setq ch (read-char nil nil nil t))
               (cond ((null ch) ;; EOF.
                      (setq *newick-interior-node-alist* nil)
                      (return-from newick-read-file (nreverse ans)))
                     ((newick-white-p ch))
                     ((eql ch #\[) (newick-consume-comment))
                     (t (return))))

; If CH is an open paren, accumulate onto ANS one Newick expression;

              (cond
               ((eql #\( ch)
                (setq *newick-interior-node-alist* nil)
                (ignore-errors
                  (setq ch (newick-cleanup (newick-read-list nil))))
                (when (or (characterp ch)
                          (not (eql #\; (read-char nil nil nil t))))
                  (newick-read-file-error-report osi ch))
                (push (cons ch *newick-interior-node-alist*) ans))

; otherwise accumulate as strings the lines since ST was last set.

               (t (let ((e (and st (file-position si))))
                    (when (and e st (file-position si st))
                      (loop (push (read-line nil t nil t) ans)
                            (if (< e (file-position si))
                                (return))))))))))))


;  HONS-LET and MEMOIZE-LET

; It might be a good enhancement to HT-LET to permit the carrying
; forward, with HOPY-CONS-CONSUME, of other honses.

(defun not-memoized-error (f)
  (ofe "~&; not-memoized-error:  ** Error:  ~a is not memoized." f))

(defmacro hons-let (form)
  (let ((old-cdr-ht     (gensym "OLD-CDR-HT"))
        (old-cdr-ht-eql (gensym "OLD-CDR-HT-EQL"))
        (old-nil-ht     (gensym "OLD-NIL-HT")))
    `(let ((,old-cdr-ht *hons-cdr-ht*)
           (,old-cdr-ht-eql *hons-cdr-ht-eql*)
           (,old-nil-ht *nil-ht*))
       (unwind-protect
           (progn (clear-hash-tables) ,form)
         (setq *hons-cdr-ht* ,old-cdr-ht)
         (setq *hons-cdr-ht-eql* ,old-cdr-ht-eql)
         (setq *nil-ht* ,old-nil-ht)))))

(defmacro memoize-let (fn form)
  (let ((fn-name (gensym "FN-NAME"))
        (tablevalue (gensym "TABLEVALUE"))
        (ponstablevalue (gensym "PONSTABLEVALUE"))
        (h (gensym "H"))
        (ht1 (gensym "HT1")))
    `(let* ((,fn-name ,fn)
            (,h (memoizedp-raw ,fn-name)))
       (unless ,h (not-memoized-error ,fn-name))
       (let* ((,tablevalue
               (symbol-value
                (access memoize-info-ht-entry ,h :tablename)))
              (,ponstablevalue
               (symbol-value
                (access memoize-info-ht-entry ,h :ponstablename)))
              (,ht1 (mht)))
         (unwind-protect
             (progn (setf (symbol-value
                           (access memoize-info-ht-entry ,h
                                   :tablename))
                          ,ht1)
                    (setf (symbol-value
                           (access memoize-info-ht-entry ,h
                                   :ponstablename))
                           (mht))
                    ,form)
           ;; During the evaluation of form, a change to a stobj may
           ;; result in tablename getting a new value, in which case
           ;; we may not restore its old value.  And a change in the
           ;; memoization status of fn would make a restoration
           ;; pointless.
           (let ((test (and (eq (symbol-value
                                 (access memoize-info-ht-entry
                                         ,h :tablename))
                                ,ht1)
                            (eq ,h (memoizedp-raw ,fn-name)))))
             (setf (symbol-value (access memoize-info-ht-entry
                                         ,h :tablename))
                   (and test ,tablevalue))
             (setf (symbol-value (access memoize-info-ht-entry
                                         ,h :ponstablename))
                   (and test ,ponstablevalue))))))))

(defmacro memoize-on (fn x)
  (let ((f (gensym "F"))
        (h (gensym "H"))
        (o (gensym "O")))
    `(let* ((,f ,fn) (,h (memoizedp-raw ,f)))
       (unless ,h (not-memoized-error ,f))
       (let ((,o (symbol-function (the symbol ,f))))
         (unwind-protect
             (progn (setf (symbol-function (the symbol ,f))
                          (symbol-value
                           (the symbol (access memoize-info-ht-entry
                                               ,h :memoized-fn))))
                    ,x)
           (setf (symbol-function (the symbol ,f)) ,o))))))

(defmacro memoize-off (fn x)
  (let ((f (gensym "F"))
        (h (gensym "H"))
        (o (gensym "O")))
    `(let* ((,f ,fn) (,h (memoizedp-raw ,f)))
       (unless ,h (not-memoized-error ,f))
       (let ((,o (symbol-function (the symbol ,f))))
         (unwind-protect
             (progn (setf (symbol-function (the symbol ,f))
                          (symbol-value
                           (the symbol (access memoize-info-ht-entry
                                               ,h :old-fn))))
                    ,x)
           (setf (symbol-function (the symbol ,f)) ,o))))))

(defun number-of-args (fn)

; For simplicity and speed, we only memoize functions that take a
; fixed number of arguments and return a fixed number of values.
; NUMBER-OF-ARGS returns an integer if the number of args of FN is
; fixed, and otherwise NIL.

  (let* ((state *the-live-state*) (w (w state)))
    (and (symbolp fn)
         (fboundp fn)
         (not (macro-function fn))
         (not (special-operator-p fn))
         (let ((x (getprop fn 'stobjs-in t 'current-acl2-world w)))
           (cond
            ((or (eq x t) (member 'state x))
             #+openmcl
             (multiple-value-bind (a b c)

; No idea whether or why this really works.  Ask Gary Byers.
; Hope this can be done right in other Lisps.

                 (ccl::min-max-actual-args (symbol-function fn) 0)
               (declare (ignore c))
               (and (integerp a) (eql a b) a))
             #-openmcl nil)
            (t (length x)))))))

(defun number-of-return-values (fn)
  (let* ((state *the-live-state*) (w (w state)))
    (let ((x (getprop fn 'stobjs-out t 'current-acl2-world w)))
      (cond ((or (eq x t) (eq x nil) (member 'state x))
             nil)
            (t (length x))))))

(defun profile (fn &key
                   (number-of-args (number-of-args fn))
                   (number-of-return-values
                    (or (number-of-return-values fn)
                        1)) ; 1 is a dumb guess.
                   (inline nil)
                   (trace nil)
                   (condition nil)
                   )
  (when (memoizedp-raw fn)
    (oft "~&; profile:  Unmemoizing ~a." fn)
    (unmemoize-fn fn))
  (cond ((member fn '(intern find-symbol))
         (memoize-fn fn
                     :stobjs-in '(nil nil)
                     :stobjs-out '(nil nil)
                     :inline inline
                     :trace trace
                     :condition condition
                     :formals '(x y)))
        ((null (number-of-args fn))
         (ofd "~&; profile: ** Error: Could not determine the ~
              number of args for ~s." fn))
        (t (memoize-fn fn
                       :condition  condition
                       :inline     inline
                       :trace trace
                       :formals
                       (loop for i below number-of-args
                             collect (intern (ofn "x-~a" i)
                                             *acl2-package*))
                       :stobjs-in  (make-list number-of-args)
                       :stobjs-out
                       (make-list number-of-return-values)))))

(defun profile-unmemoized-starting-at (n)
  (when (symbolp n) (setq n (event-number n)))
  (check-type n integer)
  (let* ((state *the-live-state*)
         (packs (get-global 'packages-created-by-defpkg state))
         (w (w state)))
    (labels
        ((no-use (pn fn)
           (let ((x (getprop fn pn t 'current-acl2-world w)))
             (or (eq x t) (member 'state x))))
         (unmemoizable (fn)
                       (or (memoizedp-raw fn)
                           (not (fboundp fn))
                           (eq (symbol-package fn)
                               *main-lisp-package*)
                           (gethash fn *memoize-reject-ht*)
                           (macro-function fn)
                           (special-form-or-op-p fn)
                           (no-use 'stobjs-in fn)
                           (no-use 'stobjs-out fn)
                           (no-use 'formals fn))))
      (loop for p in (cons "ACL2" packs) do
            (do-symbols (fn p)
              (unless (or (unmemoizable fn)
                          (null (number-of-args fn))
                          (< (event-number fn) n))
                (profile fn)))))))
  
(defun modules ()
  (strip-cars (table-alist 'defm-table (w *the-live-state*))))


; Plev0 provides some support for printing control.

(defun plev0 (level length lines tuple)
  (cond ((not (standard-evisc-tuplep tuple))
         (ofe "plev0: *** Illegal tuple: ~a." tuple)))
  (cond ((not (or (null length)
                  (and (integerp length) (>= length 0))))
         (ofe "plev0: *** A length argument to plev1 must be an ~
                 integer or nil, not ~a." length)))
  (cond ((not (or (null level) (and (integerp level) (>= level 0))))
         (ofe "plev0: *** A level argument to plev1 must be an ~
                 integer or nil, not ~a." level)))
  (cond ((not (or (null lines) (and (integerp lines) (>= lines 0))))
         (ofe "plev0: *** A lines argument to plev1 must be an ~
                 integer or nil, not ~a." lines)))
  (setq *print-length* length)
  (setq *print-level* level)
  (setq *print-lines* lines)
  #+openmcl
  (progn
    (setq ccl:*backtrace-print-level* level)
    (setq ccl:*trace-print-level* level)
    (setq ccl:*backtrace-print-length* length)
    (setq ccl:*trace-print-length* length))
  (setf ; avoid warning that *trace-evisc-tuple* is undefined
   (symbol-value '*trace-evisc-tuple*) tuple)
  (setq *approved-user-default-evisc-tuple* tuple)
  (setq *approved-user-term-evisc-tuple*    tuple)
  nil)

(defun honsp-check (x)
  (cond
   ((consp x)
    #+hons
    (when (not (honsp x))
      (er hard 'honsp-check
          "The value ~X01 is a consp but, contrary to expectation, ~
           not a honsp."
          x
          (list nil 3 4 nil)))
    t)
   ((typep x '(and array string))
    #+hons
    (when (not (gethash x *hons-str-ht*))
      (er hard 'honsp-check
          "The value ~X01 is a string or bignum but, contrary to ~
           expectation, not a string or bignum hashed for the HONS ~
           implementation."
          x
          (list nil 3 4 nil)))
    t)
   (t nil)))

(defun needs-shrinking (h)
  (let ((size (hash-table-size h)))
    (declare (fixnum size))
    (and (> size (cond ((eq h *hons-cdr-ht*) *hons-cdr-ht-size*)
                       ((eq h *nil-ht*) *nil-ht-size*)
                       (t *small-ht-size*)))
         (< (* 3 (hash-table-count h)) size))))

(defun maybe-shrink-sub (h)
  (cond ((consp h)
         (let ((nh (mht :size (length h) :weak :value)))
           (loop for x in h do (setf (gethash (car x) nh) x))
           nh))
        ((needs-shrinking h)
         (let ((nh (mht :size (* 3 (hash-table-count h))
                        :weak :value)))
           (maphash (lambda (k v) (setf (gethash k nh) v)) h)
           nh))
        (t h)))

; We sometimes replace lists with subsidiary hash tables even though
; short lists are faster to search.  By converting to hash tables, we
; permit the garbage collection of honses that are referenced by no
; one else, thanks to idea of 'weak' in OpenMCL.  We sometimes convert
; subsidiary hash tables back to lists when their counts are low
; enough because short lists are faster to search.

; Possible improvement: In maybe-shrink-..., it might be a better idea
; to move all the honses on lists to a single weak value hash table,
; instead of to many small hash tables, then, do the gc, and then
; rehash all those values (back to lists).  One downside to putting
; all those honses into a hash table is that they would all have to be
; rehashed to be placed back into the correct sub lists; if they are
; kept in separate small hashtables, one merely needs to maphash and
; create a list.

(defun maybe-shrink-main (h)
  (maphash (lambda (k v)
             (cond ((and (not (listp v))
                         (eql 0 (hash-table-count v)))
                    (remhash k h))
                   (t (let ((nv (maybe-shrink-sub v)))
                        (or (eq nv v) (setf (gethash k h) nv))))))
           h)
  (cond ((needs-shrinking h)
         (let ((nh (mht :test (hash-table-test h)
                        :size (* 3 (hash-table-count h))
                        #+openmcl :weak
                        #+openmcl (ccl::hash-table-weak-p h))))
               (maphash (lambda (k v) (setf (gethash k nh) v)) h)
               nh))
        (t h)))

(defun ht-list (h)
  (maphash
   (lambda (k v)
     (if (not (listp v))
         (let ((c (hash-table-count v)))
           (declare (fixnum c))
           (cond ((eql 0 c) (remhash k h))
                 ((< c *start-car-ht-size*)
                  (setf (gethash k h)
                        (let (l)
                          (maphash (lambda (k v) (declare (ignore k))
                                     (push v l))
                                   v)
                          l)))))))
   h))

#+openmcl
; *GC-MSG-FOR-OUR-GC* is set to T after a gc in OpenMCL.
(defglobal *gc-msg-for-our-gc* nil)
(defun our-gc ()
  #+openmcl
  (progn (setq *gc-msg-for-our-gc* nil)
         (ccl::gc)
         (loop (sleep .01)
               (when *gc-msg-for-our-gc*
                 (setq *gc-msg-for-our-gc* nil)
                 (return nil))))
  #+GCL
  (si::gbc t))

(defun maybe-shrink-some-hash-tables ()
  (setq *hons-cdr-ht*     (maybe-shrink-main *hons-cdr-ht*))
  (setq *hons-cdr-ht-eql* (maybe-shrink-main *hons-cdr-ht-eql*))
  (setq *nil-ht*     (maybe-shrink-sub  *nil-ht*))
  (our-gc)
  (ht-list *hons-cdr-ht*)
  (ht-list *hons-cdr-ht-eql*)
  (our-gc))

(defglobal *hons-readtable-init-done* nil)
(defun hons-readtable-init ()
  (setq *hons-readtable* (copy-readtable *acl2-readtable*))
  (let ((*readtable* *hons-readtable*))
    (set-macro-character #\( #'hons-read-reader)
    (set-macro-character #\) #'close-paren-read-macro)
    (set-macro-character #\. #'dot-read-macro t))
  (setq *hons-readtable-init-done* t))

(defglobal *compact-read-init-done* nil)
(defun compact-read-init ()
  (setq *compact-read-file-readtable*
        (copy-readtable *hons-readtable*))
  (let ((*readtable* *compact-read-file-readtable*))
    (set-dispatch-macro-character #\# #\# #'ns-ns-reader)
    (set-dispatch-macro-character #\# #\= #'ns-=-reader))
  (setq *compact-read-init-done* t))

(defun float-ticks/second-init ()
  (setq *float-ticks/second*
        #+RDTSC64
        (let ((i1 (ccl::rdtsc64))
              (i2 (progn (sleep .01) (ccl::rdtsc64))))
          (if (>= i2 i1)
              (* 100 (float (- i2 i1)))
            (ofe "~&; hons-init:  ** Error:  RDTSC nonsense.")))
        #-RDTSC64
        (float internal-time-units-per-second))
  (check-type *float-ticks/second*
              (and float (satisfies plusp))))

(defmacro prinl (&rest r) 
  
; PRINL prints the members of r interleaved with their values.  For
; example, (prinl a b (+ a b)) might print:
; A => 1
; B => 2
; (+ A B) => 3
  
  `(our-syntax
    (let ((*print-readably* nil))
      ,@(loop for x in r collect `(ofd "~%~s => ~s." ',x ,x))
      nil)))

; WATCH

; WATCH-FORMS may be freely redefined to produce whatever forms one
; wants watched.

(defparameter *watch-forms*
  '("Strings and quoted forms are ignored in *watch-forms*."
    (time-since-watch-start)
    (-----)
    #+openmcl
    (gc-count)
    #+openmcl
    (sh (ofn "ps -p ~a -o %cpu,%mem,vsz,rssize,start,time" (ccl::getpid)))
    '(sh "date" nil)
    '(sh "uptime" nil)
    '(current-form)
    (print-call-stack)
    (memoize-summary)
    (-----)
    (hons-calls/sec-run-time)
    (hons-hits/calls)
    (hons-acons-summary)
    (pons-calls/sec-run-time)
    (pons-hits/calls)
    '(hons-summary)
    '(run-top)
    '(room t)
    #+openmcl
    '(number-of-cpus-on-this-machine)
    '(acl2-doc 'memoize-summary)
    '(acl2-doc 'watch)
    (current-values-of-some-vars)
    ;(sh "printenv")
    '(timer-function-name)))

(defn acl2-doc (x)
  (-----)
  (fresh-line)
  (princ (get-doc-string x *the-live-state*))
  (fresh-line))

(defn run-top ()
  #+darwin-host
  (sh "top -l 1" t)
  #+linux-host
  (sh \"top -n 1\") t)

(defglobal *watch-process* nil)

(defglobal *watch-start-real-time* 0)
(defglobal *watch-start-run-time* 0)
(defglobal *watch-last-run-time* 0)

(defglobal *watch-tmp-text* "~/watch-tmp.text")
(defglobal *watch-tmp-lisp* "~/watch-tmp.lisp")

(defglobal *watch-tmp-lisp-write-date* 0)

(defglobal *watch-semaphore*
  #+openmcl
  (ccl::make-semaphore)
  #-openmcl
  nil)

(defglobal *watch-il*
  #+openmcl
  (slot-value ccl:*application*
              'ccl::initial-listener-process)
  #-openmcl
  nil)

(defglobal *watch-string*
  (make-array 20000
              :element-type 'character
              :adjustable t
              :fill-pointer t))

(defparameter *watch-vars*
  '(*watch-forms*
    *memoize-summary-show-tables*
    *memoize-debug*
    *watch-sleep-seconds*
    *watch-refresh*
    *watch-refresh-when-inactive*
    *report-bytes-used*
    *report-hons-calls*
    (*memoize-summary-order*
     total-time
     bytes-used
     hons-calls
     hits/calls 
     time/call
     number-of-calls
     event-number)
    *watch-vars*))

(defn current-values-of-some-vars ()
  (-----)
  (let ((*print-pretty* t)
        (*print-level* nil)
        (*print-length* nil)
        (*print-case* :downcase))
    (loop for v in *watch-vars* do
          (cond ((and (symbolp v) (boundp v))
                 (oft "~%(setq ~s " v)
                 (let* ((x (symbol-value v))
                        (y (if (or (symbolp x) (consp x))
                               (list 'quote x)
                             x)))
                   (if (consp x) (oft "~% "))
                   (prin1 y)
                   (oft ")")))
                ((and (consp v)
                      (symbolp (car v))
                      (boundp (car v)))
                 (oft "~%(setq ~s " (car v))
                 (let* ((x (symbol-value (car v)))
                        (y (if (or (symbolp x) (consp x))
                               (list 'quote x)
                             x)))
                 (if (consp x) (oft "~% "))
                 (prin1 y)
                 (oft ")")
                 (oft "~& Other options:~% " (car v))
                 (prin1 (cdr v)))))))
  nil)

(defun watch-kill ()
  (loop for file in (list *watch-tmp-text* *watch-tmp-lisp*)
        do (when (probe-file file) (delete-file file)))
  #+openmcl
  (if (ccl::processp *watch-process*)
      (ccl::process-kill *watch-process*))
  (setq *watch-process* nil))

; (pushnew :watch-debug *features*)
; (old-trace watch-print-forms watch-dump watch-interrupt
; watch-function)

(defglobal *watch-error* nil)

(defun watch-dump ()
  (with-open-file (s *watch-tmp-text*
                     :direction :output
                     :if-does-not-exist :create
                     :if-exists :overwrite)
    #+openmcl
    (ccl::stream-length s 0)
    #+watch-debug (prinl *watch-string*)
    (write-string *watch-string* s)
    #+openmcl
    (ccl::stream-length s (length *watch-string*))
    (setq *watch-last-run-time* (get-internal-run-time))))

(defun watch-interrupt ()
  (unwind-protect
      (our-syntax
       (let ((*print-case* :downcase))
         #+watch-debug (print "running watch-interrupt" t)
         (let ((d (file-write-date *watch-tmp-lisp*)))
           (when (and d (< *watch-tmp-lisp-write-date* d))
             (setq *watch-tmp-lisp-write-date* d)
             (load *watch-tmp-lisp* :print t)))
         #+watch-debug
         (print "running watch-print-forms" t)
         (with-output-to-string
           (*standard-output* *watch-string*)
           (setf (fill-pointer *watch-string*) 0)
           (loop for form in *watch-forms* do
                 (let* ((z (cons nil nil))
                        (val z))
                   (ignore-errors (setq val (eval form)))
                   (when (eq val z)
                     (oft "~&~a~50t? error in eval ?" form)))))
         #+watch-debug (print "running watch-dump" t)
         #+openmcl
         (watch-dump))
       #+watch-debug (print "signalling")
       #+openmcl
       (ccl::signal-semaphore *watch-semaphore*)
       nil)))

(defparameter *watch-sleep-seconds* 1)
(defparameter *watch-refresh* 5)
(defparameter *watch-refresh-when-inactive* 15)

(defn watch-function ()
  #+openmcl
  (loop for i from 0 do
        (cond ((or (let ((d (file-write-date *watch-tmp-lisp*)))
                     (and d (< *watch-tmp-lisp-write-date* d)))
                   (zerop (mod i *watch-refresh-when-inactive*))
                   (and (zerop (mod i *watch-refresh*))
                        (> (/ (- (get-internal-run-time)
                                 *watch-last-run-time*)
                              (float internal-time-units-per-second))
                           .01)))
               #+watch-debug (print "interrupting" t)
               #+openmcl
               (ccl::process-interrupt *watch-il* #'watch-interrupt)
               #+watch-debug (print "waiting" t)
               #+openmcl
               (ccl::wait-on-semaphore *watch-semaphore*)))
        (sleep *watch-sleep-seconds*)))

(defun watch ()
  #+openmcl
  (progn
    (watch-kill)
    (setq *watch-start-real-time* (get-internal-real-time))
    (setq *watch-start-run-time* (get-internal-run-time))
    (setq *watch-last-run-time* (get-internal-run-time))
    #+watch-debug (print "starting watch-process" t)
    (setf (fill-pointer *watch-string*) 0)
    (watch-dump)
    (setq *watch-semaphore* (ccl::make-semaphore))
    (setq *watch-il*
          (slot-value ccl:*application*
                      'ccl::initial-listener-process))
    (setq *watch-process*
          (ccl::process-run-function "Watch" #'watch-function)))
  *watch-value*)

;  Miscellaneous functions for optional use with WATCH.

(defun nchars (n c &optional (stream))
  (check-type n integer)
  (check-type c character)
  (loop (when (<= n 0) (return))
        (write-char c stream)
        (decf n)))

(defmacro defw (fn &rest r)
  `(defn ,fn ()
     (let ((fn ',fn))
       ,@r)))

(defw time-of-last-watch-update
  (multiple-value-bind (sec mi h d mo y)
      (decode-universal-time (get-universal-time))
    (let (m)
      (cond ((> h 12)
             (setq m " p.m.")
             (setq h (- h 12)))
            (t (setq m " a.m.")))
      (let* ((ans (ofn "~2,d:~2,'0d:~2,'0d~a ~4d/~d/~d"
                       h mi sec m y mo d)))
        (oft "~&~v<~a~;~a~>"
             *print-right-margin* fn ans)))))

(defun watch-time ()
  (/ (- (get-internal-real-time) *watch-start-real-time*)
     (float internal-time-units-per-second)))

(defun watch-run-time ()
  (/ (- (get-internal-run-time) *watch-start-run-time*)
     (float internal-time-units-per-second)))

(defw hons-calls/sec-run-time
  (let* ((c *hons-call-counter*)
         (ans
          (cond ((eql c 0) "No hons calls yet.")
                (t (ofn "~,1e" (round (/ c (watch-run-time))))))))
    (oft "~&~v<~a~;~a~>" *print-right-margin* fn ans)))

(defw pons-calls/sec-run-time
  (let* ((c *pons-call-counter*)
         (ans
          (cond ((eql c 0) "No pons calls yet.")
                (t (ofn "~,1e" (round (/ c (watch-run-time))))))))
    (oft "~&~v<~a~;~a~>" *print-right-margin* fn ans)))

(defw pons-hits/calls
  (let* ((c *pons-call-counter*)
         (h *pons-hit-counter*)
         (ans
          (cond ((eql c 0) "No pons calls yet.")
                (t (ofn "~,1e / ~,1e = ~,2f" h c (/ h c))))))
    (oft "~&~v<~a~;~a~>" *print-right-margin* fn ans)))

(defw hons-hits/calls
  (let* ((c *hons-call-counter*)
         (h (- c *hons-misses-counter*))
         (ans
          (cond ((eql c 0) "No hons calls yet.")
                (t (ofn "~,1e / ~,1e = ~,2f" h c (/ h c))))))
    (oft "~&~v<~a~;~a~>" *print-right-margin* fn ans)))

#+openmcl
(defw gc-count ()
  (let (h mi (sec (floor (ccl::gctime) 1000)))
    (multiple-value-setq (mi sec) (floor sec 60))
    (multiple-value-setq (h mi) (floor mi 60))
    (oft "~&~v,20<~a~;~a~;~a~;~2d:~2,'0d:~2,'0d~>"
         *print-right-margin*
         fn
         (cadr (multiple-value-list (ccl::gccounts)))
         "time in gc "
         h mi sec)))

#+openmcl
(defw number-of-cpus-on-this-machine
  (let* ((ans (ofn "~s" (ccl::cpu-count))))
    (oft "~&~v<~a~;~a~>" *print-right-margin* fn
         #+openmcl ans
         #-openmcl "Unknown")))

(defun sh (command-string &optional (newline t) (print-command nil))
  #-openmcl (declare (ignore command-string newline print-command))
  #+openmcl
  (let ((file (ofn "/tmp/watch-~a.XXXX" (ccl::getpid))))
    (ccl::os-command (ofn "~a >& ~a" command-string file))
    (when newline
      (-----)
      (fresh-line)
      (if print-command (princ command-string))
      (fresh-line)
      (with-open-file (stream file)
        (let (c)
          (loop while (setq c (read-char stream nil nil nil))
                do (write-char c)))))
    (when (not newline)
      (let ((str (with-output-to-string
                   (*standard-output*)
                   (with-open-file
                     (stream file)
                     (let (c)
                       (loop while (setq c (read-char stream nil nil nil))
                             do (write-char c)))))))
        (oft "~v<~a~;~a~>"
             *print-right-margin*
             (if print-command command-string "")
             str)))))

(defun sh-to-string (command-string)
  #-openmcl (declare (ignore command-string))
  #+openmcl
  (let ((file (ofn "/tmp/watch-~a.XXXX" (ccl::getpid))))
    (ccl::os-command (ofn "~a >& ~a" command-string file))
    (fresh-line)
    (princ command-string)
    (with-output-to-string (s)
      (with-open-file (stream file)
        (let (c)
          (terpri)
          (loop while (setq c (read-char stream nil nil nil))
                do (write-char c s))))))
  nil)

(defun ----- ()
  (fresh-line)
  (nchars
   (or 
    *print-right-margin*
    72)
   #\-))

(defun current-form ()
  (-----)
  (fresh-line)
  (princ "current-form")
  (fresh-line)
  (write - :level 3 :length 3 :case :downcase
           :readably nil))

(defw time-since-watch-start ()
  (let (h1 mi1
          (sec1 (round (watch-time)))
          h2 mi2
          (sec2 (round (watch-run-time))))
    (multiple-value-setq (mi1 sec1) (floor sec1 60))
    (multiple-value-setq (h1 mi1) (floor mi1 60))
    (multiple-value-setq (mi2 sec2) (floor sec2 60))
    (multiple-value-setq (h2 mi2) (floor mi2 60))
    (oft "~&~v<~a~;~2d:~2,'0d:~2,'0d~;~a~;~2d:~2,'0d:~2,'0d~>"
         *print-right-margin*
         fn h1 mi1 sec1
         "run-time" h2 mi2 sec2)))

(defw timer-function-name
  (let ((ans
         #+RDTSC64
         "OpenMCL's CCL::RDTSC64 function."
         #-RDTSC64
         "Common Lisp's GET-INTERNAL-REAL-TIME function."))
    (oft "~&~v<~a~;~a~>" *print-right-margin* fn ans)))

; Optional older and shorter names for a few commands.

(defun memsum (&rest r)
  (apply #'memoize-summary r))
(defun hsum (&rest r)
  (apply #'hons-summary r))
(defun memstat (&rest r)
  (apply #'memoize-statistics r))
(defun hstat (&rest r)
  (apply #'hons-statistics r))
(defmacro memo-on (&rest r)
  `(memoize-on ,@r))
(defmacro memo-off (&rest r)
  `(memoize-off ,@r))
(defun clear-memo-tables (&rest r)
  (apply #'clear-memoize-tables r))
(defun clear-hash-and-memo-tables (&rest r)
  (apply #'clear-hash-and-memoize-tables r))

; Matt, I have no idea which evisc tuple to use.

; prine stands for something like print eviscerated.

(defun prine-tuple ()
  (if (f-boundp-global 'user-default-evisc-tuple *the-live-state*)
      (f-get-global 'user-default-evisc-tuple *the-live-state*)
    (term-evisc-tuple nil *the-live-state*)))
    
(defun remove-evisc-marks (x)
  (cond ((atom x) x)
        ((eq (car x) :evisceration-mark) (cdr x))
        (t (cons (remove-evisc-marks (car x))
                 (remove-evisc-marks (cdr x))))))

(defun remove-evisc-marks-al (o)
  (let ((tuple (prine-tuple)))
    (loop for p in o collect
          (list (car p)
                '=
                (remove-evisc-marks
                 (eviscerate
                  (cdr p) (cadr tuple) (caddr tuple)
                  (car tuple) (cadddr tuple)
                  nil ; ? a guess
                  ))))))

(defun prine (obj &optional stream)
  (let ((tuple (prine-tuple)))
    (princ (remove-evisc-marks
            (eviscerate
             obj
             (cadr tuple)
             (caddr tuple)
             (car tuple)
             (cadddr tuple)
             nil ; ? a guess
             ))
           stream))
  obj)

(defun prine-list (obj &optional stream)
  (cond ((atom obj)
         (princ obj stream))
        (t (write-char #\( stream)
           (loop for tail on obj do
                 (prine (car tail) stream)
                 (cond ((consp (cdr tail))
                        (write-char #\Space stream))
                       ((null (cdr tail))
                        (write-char #\) stream))
                       (t (write-char #\Space stream)
                          (princ (cdr tail) stream)
                          (write-char #\) stream))))))
  obj)

(defglobal *mapped-octet-vectors* nil)

(defun file-vector (fn)
  #-openmcl (declare (ignore fn))
  #+openmcl
  (let ((x (ccl::map-file-to-octet-vector fn)))
    (push x *mapped-octet-vectors*)
    (array-displacement x))
  nil)

(defun set-gc-threshold (bound)
  #-openmcl (declare (ignore bound))
  #+openmcl
  (when (< (ccl::lisp-heap-gc-threshold) bound)
    (ofd "~&; Setting (ccl::lisp-heap-gc-threshold) to ~:d bytes."
         bound)
    (ccl::set-lisp-heap-gc-threshold bound)
    (ccl::use-lisp-heap-gc-threshold))
  nil)

#+openmcl
(defun maybe-set-gc-threshold (&optional (fraction 1/32))
  ; Jared Davis suggested 1/32.
  (let (n)
    (ignore-errors (setq n (physical-memory)))
    (cond ((and (integerp n) (> n (* 2 (expt 10 9))))
           (setq n (floor (* n fraction)))
           (set-gc-threshold n)))))

(defglobal *hons-init-hook*
  '(progn
     #+openmcl
     (progn
       ;; Subtle things that a user might want to put into
       ;; ~/openmcl-init.lisp but may not know to do so.
       (unless (equal '(t t) (multiple-value-list (ccl::gc-verbose-p)))
         ;; It seems important to watch what the gc is doing.
         (oftr "~&; Setting OpenMCL's gc verbose.")
         (ccl::gc-verbose t t))
       (when (ccl::egc-active-p)
         ;; OpenMCL's ephemeral gc doesn't work well with
         ;; honsing and memoizing, it seems.
         (oftr "~&; Turning off OpenMCL's ephemeral gc.")
         (ccl::egc nil))
       ;; Allocate heap space perhaps a little more generously
       ;; than OpenMCL's default.
       (maybe-set-gc-threshold)
       (when (and (boundp 'ccl::*quit-on-eof*)
                  (not (eq t (symbol-value 'ccl::*quit-on-eof*))))
         (oftr "~&; Control-d exits.")
         (setf (symbol-value 'ccl::*quit-on-eof*) t)))
     nil))
     
(defun hons-init ()

; HONS-INIT may be repeatedly called, so choose things to repeat
; carefully.

  (f-put-global 'evisc-print-lines nil *the-live-state*)
  (eval *hons-init-hook*)
  (unless *init-hash-tables-done* (init-hash-tables))
  (unless *hons-readtable-init-done* (hons-readtable-init))
  (unless *memoize-init-done* (memoize-init))
  (unless *compact-read-init-done* (compact-read-init))
  (float-ticks/second-init) ; Good idea to repeat.
  #+openmcl
  (ccl::add-gc-hook
   (lambda ()
     (setq *gc-msg-for-our-gc* t)
     (setq *last-gc-exit-time* (internal-real-time)))
   :post-gc)
  #+openmcl
  nil)
